Package: rosm
Type: Package
Title: Plot Raster Map Tiles from Open Street Map and Other Sources
Version: 0.3.1
Encoding: UTF-8
Authors@R: c(person("Dewey", "Dunnington", role = c("aut", "cre"),
				email = "dewey@fishandwhistle.net", comment = c(ORCID = "0000-0002-9415-4582")),
			person("Timothée", "Giraud", role = "ctb"))
Maintainer: Dewey Dunnington <dewey@fishandwhistle.net>
Description: Download and plot Open Street Map <https://www.openstreetmap.org/>,
    Bing Maps <https://www.bing.com/maps> and other tiled map sources. Use to create 
    basemaps quickly and add hillshade to vector-based maps.
License: GPL-2
Imports: curl, jpeg, png, wk, glue, progress, rlang
Suggests: sp, plyr, raster, testthat (>= 3.0.0), withr, sf, terra,
        abind, methods, jsonlite, tiff, vdiffr
URL: https://github.com/paleolimbot/rosm
BugReports: https://github.com/paleolimbot/rosm/issues
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-01-21 04:13:41 UTC; deweydunnington
Author: Dewey Dunnington [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9415-4582>),
  Timothée Giraud [ctb]
Repository: CRAN
Date/Publication: 2026-01-21 06:40:23 UTC
Built: R 4.6.0; ; 2026-04-25 23:29:38 UTC; unix

Package: rosm
Type: Package
Title: Plot Raster Map Tiles from Open Street Map and Other Sources
Version: 0.3.1
Encoding: UTF-8
Authors@R: c(person("Dewey", "Dunnington", role = c("aut", "cre"),
				email = "dewey@fishandwhistle.net", comment = c(ORCID = "0000-0002-9415-4582")),
			person("Timothée", "Giraud", role = "ctb"))
Maintainer: Dewey Dunnington <dewey@fishandwhistle.net>
Description: Download and plot Open Street Map <https://www.openstreetmap.org/>,
    Bing Maps <https://www.bing.com/maps> and other tiled map sources. Use to create 
    basemaps quickly and add hillshade to vector-based maps.
License: GPL-2
Imports: curl, jpeg, png, wk, glue, progress, rlang
Suggests: sp, plyr, raster, testthat (>= 3.0.0), withr, sf, terra,
        abind, methods, jsonlite, tiff, vdiffr
URL: https://github.com/paleolimbot/rosm
BugReports: https://github.com/paleolimbot/rosm/issues
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-01-21 04:13:41 UTC; deweydunnington
Author: Dewey Dunnington [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9415-4582>),
  Timothée Giraud [ctb]
Repository: CRAN
Date/Publication: 2026-01-21 06:40:23 UTC
Built: R 4.6.0; ; 2026-04-28 13:43:51 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as_osm_url_spec,character)
S3method(as_osm_url_spec,osm_url_spec)
export(as.tile_source)
export(as_osm_url_spec)
export(bmaps.plot)
export(bmaps.types)
export(extract_bbox)
export(get_default_cachedir)
export(get_default_tile_source)
export(has_internet)
export(is.tile_source)
export(makebbox)
export(osm.image)
export(osm.lines)
export(osm.plot)
export(osm.points)
export(osm.polygon)
export(osm.raster)
export(osm.segments)
export(osm.text)
export(osm.types)
export(osm_crs_native)
export(osm_ensure_lnglat)
export(osm_ensure_native)
export(osm_lnglat)
export(osm_native)
export(osm_raster)
export(osm_tile)
export(osm_tile_covering)
export(osm_tile_envelope)
export(osm_tile_quadkey)
export(osm_tile_top_left)
export(osm_url)
export(osm_url_load_async)
export(osm_url_spec)
export(osm_url_spec_example)
export(osm_zoom_num_tiles)
export(register_tile_source)
export(set_default_cachedir)
export(set_default_tile_source)
export(source_from_url_format)
export(zoombbox)

ROSM: Open Street Map tiles in R
================

<!-- README.md is generated from README.Rmd. Please edit that file -->
<!-- badges: start -->

[![Codecov test
coverage](https://codecov.io/gh/paleolimbot/rosm/branch/master/graph/badge.svg)](https://app.codecov.io/gh/paleolimbot/rosm?branch=master)
<!-- badges: end -->

Download and plot [Open Street Map](https://www.openstreetmap.org/),
[Bing Maps](https://www.bing.com/maps), and other tiled map sources. A
previous version of this package provided an API for plotting base maps;
the new API is more low-level and covers tile math, URL generation, and
fetching tiles in parallel.

## Installation

The **rosm** package is [available on
CRAN](https://cran.r-project.org/package=rosm), and can be installed
using `install.packages("rosm")`.

## Example

``` r
library(rosm)

bounds <- wk::rct(
  -7476083, 5349058,
  -6594103, 6243203,
  crs = osm_crs_native()
)

(grd <- osm_raster(bounds, osm_url_spec_example()))
#> <wk_grd_rct [768 x 768] => [-7514066 5322463 -6574807 6261721] with crs=EPSG:3857>
#> List of 2
#>  $ data: 'nativeRaster' int [1:768, 1:768] -2108502 -2108502 -2108502 -2108502 -2108502 -2108502 -2108502 -2108502 -2108502 -2108502 ...
#>  $ bbox: wk_rct[1:1] [-7514066 5322463 -6574807 6261721]
#>  - attr(*, "class")= chr [1:2] "wk_grd_rct" "wk_grd"
plot(grd)
```

<img src="man/figures/README-example-1.png" width="100%" />

#' Check for Internet
#'
#' Used to skip tests and examples for this package when offline.
#'
#' @return TRUE if the internet is available, false otherwise
#' @export
#'
#' @examples
#' has_internet()
#'
has_internet <- function() {
  !is.null(curl::nslookup("r-project.org", error = FALSE))
}

#' Deprecated interface
#'
#' The previous interface for rosm was written to support idioms that are
#' no longer prevalent in modren r-spatial code. These functions may continue
#' to exist; however, their use is not encouraged and the functions may be
#' removed in a future release.
#'
#' @param x,y,x0,y0,x1,y1,url_format,max_zoom,min_zoom,attribution,extension Deprecated
#' @param key,tolatlon,epsg,toepsg,labels,n,e,s,w,factor,offset Deprecated
#' @param bbox A bounding box as generated by \code{sp::bbox()}
#' @param zoomin The amount by which to adjust the automatically calculated zoom (or
#' manually specified if the \code{zoom} parameter is passed). Use +1 to zoom in, or -1 to zoom out.
#' @param zoom Manually specify the zoom level (not reccomended; adjust \code{zoomin} or
#' \code{res} instead.
#' @param type A map type; one of that returned by \link{osm.types}. User defined types are possible
#' by defining \code{tile.url.TYPENAME <- function(xtile, ytile, zoom){}} and passing TYPENAME
#' as the \code{type} argument.
#' @param forcedownload \code{TRUE} if cached tiles should be re-downloaded. Useful if
#' some tiles are corrupted.
#' @param stoponlargerequest By default \code{osm.plot} will only load 32 tiles at a time. If
#' plotting at a higher resolution it may be necessary to pass \code{true} here.
#' @param fusetiles \code{TRUE} if tiles should be fused into a single image. This is the
#' default because white lines appear between tiles if it is set to \code{FALSE}. PDFs
#' appear not to have this problem, so when plotting large, high resolution PDFs it may be
#' faster (and more memory efficient) to use \code{fusetiles=FALSE}.
#' @param cachedir The directory in which tiles should be cached. Defaults to \code{getwd()/rosm.cache}.
#' @param res The resolution used to calculate scale.
#' @param project \code{TRUE} if tiles should be projected to a pseudo-mercator projection,
#' \code{FALSE} if lat/lon should be maintained. Becuase \code{sp::plot} adjusts the aspect
#' according to latitude for lat/lon coordinates, this makes little difference at high
#' zoom and may make plotting overlays more convenient. Defaults to \code{TRUE}.
#' @param progress A progress bar to use, or "none" to suppress progress updates
#' @param quiet Pass \code{FALSE} to see more error messages, particularly if
#'   your tiles do not download/load properly.
#' @param projection A map projection in which to reproject the RasterStack as
#'   generated by \code{CRS()} or \code{Spatial*@@proj4string}. If a
#'   \code{Spatial*} object is passed as the first argument, this argument will
#'   be ignored.
#' @param crop \code{TRUE} if results should be cropped to the specified
#'   bounding box (see \code{x}), \code{FALSE} otherwise.
#' @param filename A filename to which the raster should be written (see
#'   \code{raster::writeRaster()}). Use a ".tif" extension to write as a
#'   GeoTIFF.
#' @param resample One of "ngb" (nearest neighbour) or "bilinear". Passed to
#'   \link[raster]{projectRaster}.
#' @param api_key API key used for CARTO map types. If `NULL`, the value of
#'   the `CARTO_API_KEY` environment variable is used.
#' @param ... Arguments passed to other methods
#'
#' @rdname deprecated
#' @export
as.tile_source <- function(x, ..., api_key = NULL) {
  # first check if x is a tile source
  if (is.tile_source(x)) {
    return(x)
  }

  # order is registered -> built_in -> functions -> string format
  if ((length(x) == 1) && (x %in% names(registered_sources))) {
    src <- registered_sources[[x]]
    src$name <- x
    src
  } else if ((length(x) == 1) && (x %in% names(tile_sources))) {
    src <- tile_sources[[x]]
    if (x %in% c("cartodark", "cartolight")) {
      if (is.null(api_key)) {
        api_key <- Sys.getenv("CARTO_API_KEY", unset = NA_character_)
      }

      if (!is.character(api_key) || length(api_key) != 1 || is.na(api_key) || !nzchar(api_key)) {
        stop(
          "CARTO map types require an API key. Supply `api_key` or set ",
          "the CARTO_API_KEY environment variable.",
          call. = FALSE
        )
      }

      src <- source_from_url_format(
        paste0(src$url_formats, "?key=", curl::curl_escape(api_key)),
        attribution = src$get_attribution(),
        extension = src$get_extension()
      )
    }
    src$name <- x
    src
  } else if ((length(x) == 1) && is.character(x)) {
    # if url function exists, use old-style function definitions
    if (exists(paste0("tile.url.", x))) {
      source_from_global_functions(x)
    } else {
      # create using string format
      source_from_url_format(x, ...)
    }
  } else if ((length(x) > 1) && is.character(x)) {
    # character vectors > 1 can't be functions
    source_from_url_format(x, ...)
  } else {
    stop("Don't know how to create a tile_source from type ", class(x))
  }
}

#' @rdname deprecated
#' @export
is.tile_source <- function(x) {
  inherits(x, "tile_source")
}

#' @rdname deprecated
#' @export
source_from_url_format <- function(url_format, max_zoom = tile.maxzoom.default(),
                                   min_zoom = 0, attribution = NULL, extension = tools::file_ext(url_format[1]),
                                   ...) {
  # url format like this: https://tiles.wmflabs.org/bw-mapnik/${z}/${x}/${y}.png
  # ${q} for quadkey

  # check format: need ${z}/${x}/${y} xor ${q} (not both)
  # url_format may be a vector of names, so wrap in all()
  if (xor(
    !all(grepl("${q}", url_format, fixed = TRUE)),
    !(!all(grepl("${z}", url_format, fixed = TRUE)) &&
      !all(grepl("${x}", url_format, fixed = TRUE)) &&
      !all(grepl("${y}", url_format, fixed = TRUE)))
  )) {
    stop("url_format must contain ${q} xor ${z} and ${x} and ${y}")
  }

  # check extension
  extension <- match.arg(extension, c("png", "jpeg", "jpg"))

  # force args, since they will be used in closures
  force(max_zoom)
  force(attribution)
  extra_args <- list(...)
  force(extension)

  create_tile_source(
    get_tile_url = function(xtile, ytile, zoom, quadkey = "") {
      withx <- gsub("${x}", xtile, sample(url_format, 1), fixed = TRUE)
      withy <- gsub("${y}", ytile, withx, fixed = TRUE)
      withz <- gsub("${z}", zoom, withy, fixed = TRUE)
      # return with quadkey
      gsub("${q}", quadkey, withz, fixed = TRUE)
    },
    get_attribution = function() attribution,
    get_max_zoom = function() max_zoom,
    get_min_zoom = function() min_zoom,
    get_extension = function() extension,
    name = url_format[1],
    url_formats = url_format,
    ...
  )
}

# these are old-style custom tile naming functions
source_from_global_functions <- function(type) {
  message("Using functions as custom tile sources is deprecated. Use string formats instead.")

  # get_url is mandatory, so use match.fun
  get_url <- match.fun(paste0("tile.url.", type))

  # maxzoom function is not manditory
  if (exists(paste0("tile.maxzoom.", type))) {
    get_max_zoom <- match.fun(paste0("tile.maxzoom.", type))
  } else {
    get_max_zoom <- tile.maxzoom.default
  }

  # attribute function is not manditory
  if (exists(paste0("tile.attribute.", type))) {
    attribution <- match.fun(paste0("tile.attribute", type))
  } else {
    attribution <- function() NULL
  }

  # extension should be from tile url
  extension <- tools::file_ext(get_url(0, 0, 0)[1])

  # return tile source
  create_tile_source(
    get_tile_url = get_url,
    get_max_zoom = get_max_zoom,
    get_min_zoom = tile.minzoom.default,
    get_attribution = attribution,
    get_extension = function() extension,
    name = type
  )
}

#' @rdname deprecated
#' @export
register_tile_source <- function(...) {
  sources <- list(...)
  if (is.null(names(sources))) stop("register_source must be called with named arguments")
  if (any(nchar(names(sources)) == 0)) stop("register_source must be called with named arguments")

  # lapply as.tile_source and copy to registered_sources
  list2env(lapply(sources, as.tile_source), registered_sources)

  invisible(NULL)
}

#' @rdname deprecated
#' @export
set_default_tile_source <- function(x, ...) {
  ts <- as.tile_source(x, ...)
  registered_sources$.defaultsource <- ts
}

#' @rdname deprecated
#' @export
get_default_tile_source <- function() {
  if (exists(".defaultsource", where = registered_sources)) {
    registered_sources$.defaultsource
  } else {
    as.tile_source("osm")
  }
}

# use an environment to keep track of registered sources
registered_sources <- new.env(parent = emptyenv())

#' @rdname deprecated
#' @export
osm.types <- function() {
  c(names(tile_sources), setdiff(names(registered_sources), ".defaultsource"))
}

# base function to create specs
create_tile_source <- function(get_tile_url, get_max_zoom, get_min_zoom,
                               get_attribution, get_extension, ..., validate = TRUE) {
  # here, get_tile_url, get_max_zoom, and get_attribution are all functions
  # that get passed xtile, ytile, zoom, and quadkey (if quadkey is in the formals)
  # this is for backwards compatiblity with older tile.url.TYPE functions

  # force functions
  force(get_tile_url)
  force(get_max_zoom)
  force(get_min_zoom)
  force(get_attribution)
  force(get_extension)

  if (validate) {
    # validate the functions

    # check that get_tile_url returns a character vector of length 1
    if ("quadkey" %in% names(formals(get_tile_url))) {
      url <- get_tile_url(0, 0, 0, quadkey = "0")
    } else {
      url <- get_tile_url(0, 0, 0)
    }
    if (!is.character(url)) stop("get_tile_url must return type 'character'")
    if (length(url) != 1) stop("get_tile_url must return a vector of length 1")

    # check that maxzoom is an integer
    maxzoom <- get_max_zoom()
    if ((maxzoom %% 1) != 0) stop("get_max_zoom must return an integer")

    # check that minzoom is an integer
    minzoom <- get_min_zoom()
    if ((minzoom %% 1) != 0) stop("get_min_zoom must return an integer")

    # check that get_attribution returns a character vector of length 1
    attribution <- get_attribution()
    if (!is.null(attribution)) {
      if (!is.character(attribution)) stop("get_attribution must return a character vector")
      if (length(attribution) != 1) stop("get_attribution must return a vector of length 1")
    }

    # check that get_extension returns a character vector of length 1
    extension <- get_extension()
    if (!is.null(extension)) {
      if (!is.character(extension)) stop("get_extension must return a character vector")
      if (length(extension) != 1) stop("get_extension must return a vector of length 1")
    }
  }

  # return list of class "tile_source"
  structure(list(
    get_tile_url = get_tile_url,
    get_attribution = get_attribution,
    get_max_zoom = get_max_zoom,
    get_min_zoom = get_min_zoom,
    get_extension = get_extension,
    ...
  ), class = "tile_source")
}

# modified from http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#R

bmaps.quadkey <- function(tilex, tiley, zoom) {
  nzoom <- 2^zoom
  if (tilex < 0 || tilex >= nzoom) stop("xtile out of range: ", tilex)
  if (tiley < 0 || tiley >= nzoom) stop("ytile out of range: ", tilex)
  out <- ""
  keymap <- matrix(0:3, byrow = TRUE, ncol = 2)
  decx <- tilex / nzoom
  decy <- tiley / nzoom
  for (i in 1:zoom) {
    n <- 2^i
    x <- floor(decx * 2^i) - floor(decx * 2^(i - 1)) * 2
    y <- floor(decy * 2^i) - floor(decy * 2^(i - 1)) * 2
    out <- paste0(out, keymap[y + 1, x + 1])
  }
  out
}

tiles.bybbox <- function(bbox, zoom, epsg = 4326) {
  nwlatlon <- .tolatlon(bbox[1, 1], bbox[2, 2], epsg)
  selatlon <- .tolatlon(bbox[1, 2], bbox[2, 1], epsg)

  if (nwlatlon[1] < -180) { # fixes wraparound problem with project=F
    nwlatlon[1] <- nwlatlon[1] + 360
  }

  if (nwlatlon[1] > selatlon[1]) {
    # wrapping around backside of earth
    backsidebbox <- matrix(c(nwlatlon[1], selatlon[2], 180, nwlatlon[2]), ncol = 2, byrow = FALSE)
    backsidetiles <- tiles.bybbox(backsidebbox, zoom, 4326)
    nwlatlon[1] <- -180
  } else {
    backsidetiles <- NULL
  }

  nw <- tile.xy(nwlatlon[1], nwlatlon[2], zoom)
  se <- tile.xy(selatlon[1], selatlon[2], zoom)
  tiles <- expand.grid(x = nw[1]:se[1], y = nw[2]:se[2])

  if (is.null(backsidetiles)) {
    tiles
  } else {
    rbind(backsidetiles, tiles)
  }
}

tile.xy <- function(x, y, zoom, epsg = 4326) {
  latlon <- .tolatlon(x, y, epsg)

  lat_rad <- latlon[2] * pi / 180
  n <- 2.0^zoom
  xtile <- floor((latlon[1] + 180.0) / 360.0 * n)
  ytile <- floor((1.0 - log(tan(lat_rad) + (1 / cos(lat_rad))) / pi) / 2.0 * n)
  xtile <- max(0, min(xtile, n - 1)) # limit x, y tiles to valid tile numbers
  ytile <- max(0, min(ytile, n - 1))
  c(xtile, ytile)
}

tile.nw <- function(xtile, ytile, zoom, epsg = 4326) {
  n <- 2.0^zoom
  lon_deg <- xtile / n * 360.0 - 180.0
  lat_rad <- atan(sinh(pi * (1 - 2 * ytile / n)))
  lat_deg <- lat_rad * 180 / pi

  projected <- .fromlatlon(lon_deg, lat_deg, epsg)
  # added to support wrap around tiles
  if (lon_deg < -180 && projected[1] > 0) {
    # wrap around situation

    maxx <- .fromlatlon(180, lat_deg, epsg)[1]
    projected[1] <- projected[1] - maxx * 2
  }
  projected
}

tile.bbox <- function(xtile, ytile, zoom, epsg = 4326) {
  nw <- tile.nw(xtile, ytile, zoom, epsg)
  se <- tile.nw(xtile + 1, ytile + 1, zoom, epsg)
  matrix(c(nw[1], se[2], se[1], nw[2]),
    ncol = 2,
    byrow = FALSE, dimnames = list(c("x", "y"), c("min", "max"))
  )
}

tile.url <- function(xtile, ytile, zoom, type) {
  fun <- type$get_tile_url
  if ("quadkey" %in% names(formals(fun))) {
    fun(xtile, ytile, zoom, quadkey = bmaps.quadkey(xtile, ytile, zoom))
  } else {
    fun(xtile, ytile, zoom)
  }
}

tile.ext <- function(type) {
  type$get_extension()
}

tile.maxzoom <- function(type) {
  type$get_max_zoom()
}

tile.maxzoom.default <- function() {
  return(19)
}

tile.minzoom.default <- function() {
  return(0)
}

tile.attribute <- function(type) {
  attribution <- type$get_attribution()
  if (!is.null(attribution)) message(attribution)
}

tile.cachename <- function(xtile, ytile, zoom, type, cachedir = NULL) {
  folder <- tile.cachedir(type, cachedir)
  ext <- tile.ext(type)
  file.path(folder, paste0(zoom, "_", xtile, "_", ytile, ".", ext))
}

tile.download <- function(tiles, zoom, type = "osm", forcedownload = FALSE, cachedir = NULL,
                          progress = "text", quiet = TRUE, pause = 0.1) {
  if (!forcedownload) {
    # check which tiles exist
    texists <- vapply(1:nrow(tiles), function(i) {
      cachename <- tile.cachename(tiles[i, 1], tiles[i, 2], zoom, type, cachedir)
      return(file.exists(cachename))
    }, logical(1))
    tiles <- tiles[!texists, , drop = FALSE]
  }

  if (nrow(tiles) > 0) {
    if (progress != "none") {
      message("Fetching ", nrow(tiles), " missing tiles")
      pb <- utils::txtProgressBar(min = 0, max = nrow(tiles), width = 20, file = stderr())
    }

    tile.apply(tiles, fun = function(xtile, ytile) {
      cachename <- tile.cachename(xtile, ytile, zoom, type, cachedir)
      url <- tile.url(xtile, ytile, zoom, type)
      # pause to avoid overwhelming servers
      if (pause > 0) Sys.sleep(pause)

      if (!quiet) message("trying URL: ", url)

      # try to download file
      result <- try(curl::curl_download(url, cachename, quiet = quiet, mode = "wb"),
        silent = quiet
      )

      # display errors only in progress mode
      if (!quiet && inherits(result, "try-error")) {
        message(sprintf(
          "Failed to download tile %s:(%s, %s) for type %s / %s",
          zoom, xtile, ytile, type$name, result
        ))
      } else if (!quiet && !file.exists(cachename)) {
        message(sprintf(
          "Failed to download tile %s:(%s, %s) for type %s",
          zoom, xtile, ytile, type$name
        ))
      }

      # delete files for try-errors
      if (inherits(result, "try-error")) {
        try(unlink(cachename), silent = TRUE)
      }
    }, progress = progress)

    if (progress != "none") message("...complete!")
  }
}

# use an internal environment to cache REST information
bing_rest_queries <- new.env(parent = emptyenv())

bmaps.restquery <- function(bingtype, key = NULL) {
  # http://dev.virtualearth.net/REST/v1/Imagery/Metadata/Aerial?key=KEY
  # get a key at https://msdn.microsoft.com/en-us/library/ff428642.aspx

  # use cached information first
  if (bingtype %in% names(bing_rest_queries)) {
    result <- bing_rest_queries[[bingtype]]
  } else {
    if (is.null(key)) {
      key <- "Aut49nhp5_Twwf_5RHF6wSGk7sEzpcSA__niIXCHowQZLMeC-m8cdy7EmZd2r7Gs"
    }
    urlstring <- paste0("http://dev.virtualearth.net/REST/v1/Imagery/Metadata/", bingtype, "?key=", key)

    connect <- curl::curl(urlstring)
    lines <- try(readLines(connect, warn = FALSE), silent = TRUE)
    close(connect)

    if (inherits(lines, "try-error")) stop("  Bing REST query failed for type: ", bingtype)

    # convert to a list
    result <- jsonlite::fromJSON(paste(lines, collapse = ""), simplifyVector = FALSE)
    # cache the result
    bing_rest_queries[[bingtype]] <- result
  }
  # display the copyright notice
  message(result$copyright)

  # return only the relevant node
  result$resourceSets[[1]]$resources[[1]]
}

bmaps.sourcefromrest <- function(rest, name) {
  force(rest)
  force(name)
  create_tile_source(
    get_tile_url = function(xtile, ytile, zoom, quadkey) {
      gsub("{quadkey}", quadkey, gsub("{subdomain}", sample(rest$imageUrlSubdomains, 1),
        rest$imageUrl,
        fixed = TRUE
      ),
      fixed = TRUE
      )
    },
    get_max_zoom = function() rest$zoomMax,
    get_min_zoom = function() rest$zoomMin,
    get_attribution = function() NULL,
    get_extension = function() {
      tools::file_ext(gsub("\\?.*$", "", rest$imageUrl[1]))
    },
    name = name
  )
}

#' @rdname deprecated
#' @export
bmaps.types <- function() {
  c("Aerial", "AerialWithLabels", "Road")
}

#' @rdname deprecated
#' @export
bmaps.plot <- function(bbox, type = "Aerial", key = NULL, ...) {
  if (!(type %in% bmaps.types())) stop("type must be one of Aerial, AerialWithLabels, or Road")

  # get REST information
  rest <- bmaps.restquery(type, key)
  # plot using OSM.plot
  osm.plot(bbox = bbox, type = bmaps.sourcefromrest(rest, paste0("bing_", type)), ...)
  # plot the little bing logo
  extraargs <- list(...)
  bmaps.attribute(res = extraargs$res, cachedir = extraargs$cachedir)
}


bmaps.attribute <- function(padin = c(0.05, 0.05), res = NULL, cachedir = NULL, scale = 0.7) {
  if (is.null(res)) {
    res <- 80
  }
  # http://dev.virtualearth.net/Branding/logo_powered_by.png
  bingfile <- file.path(tile.cachedir(list(name = "bing")), "bing.png")
  if (!file.exists(bingfile)) {
    curl::curl_download("http://dev.virtualearth.net/Branding/logo_powered_by.png",
      bingfile,
      quiet = TRUE, mode = "wb"
    )
  }
  binglogo <- png::readPNG(bingfile)
  ext <- graphics::par("usr")
  rightin <- graphics::grconvertX(ext[2], from = "user", to = "inches")
  bottomin <- graphics::grconvertY(ext[3], from = "user", to = "inches")
  widthin <- dim(binglogo)[2] / res * scale
  heightin <- dim(binglogo)[1] / res * scale
  leftusr <- graphics::grconvertX(rightin - padin[1] - widthin, from = "inches", to = "user")
  bottomusr <- graphics::grconvertY(bottomin + padin[2], from = "inches", to = "user")
  topusr <- graphics::grconvertY(bottomin + padin[2] + heightin, from = "inches", to = "user")
  rightusr <- graphics::grconvertX(rightin - padin[1], from = "inches", to = "user")

  graphics::rasterImage(binglogo, leftusr, bottomusr, rightusr, topusr)
}


#' Set/Get the Default Tile Cache Location
#'
#' The default tile cache location is the "rosm.cache" folder in the current
#' working directory, but for a variety of reasons it may be desirable to
#' use one cache directory for all calls in a script. This must be called
#' every time the namespace is loaded.
#'
#' @param cachedir A path to use as the cache directory (relative to the working directory).
#'   Use NULL to reset to the default.
#'
#' @return The previous cache directory, invisibly.
#' @export
#'
#' @examples
#' set_default_cachedir(tempfile())
#' get_default_cachedir()
#' (set_default_cachedir(NULL))
#'
set_default_cachedir <- function(cachedir) {
  # NULL sets back to the default
  if (is.null(cachedir)) {
    cachedir <- "rosm.cache"
  }
  if (!is.character(cachedir) || (length(cachedir) != 1)) {
    stop("'cachedir' must be a character vector of length 1")
  }
  # keep ref to old cachedir
  old_cachedir <- rosm_state$default_cachedir
  # set new cachedir
  rosm_state$default_cachedir <- cachedir
  # return old cachedir
  invisible(old_cachedir)
}


#' @rdname set_default_cachedir
#' @export
get_default_cachedir <- function() {
  rosm_state$default_cachedir
}

rosm_state <- new.env(parent = emptyenv())
rosm_state$default_cachedir <- "rosm.cache"
# functions used by both google and osm

tile.cachedir <- function(type, cachedir = NULL) {
  if (is.null(cachedir)) {
    cachedir <- get_default_cachedir()
  }
  safename <- gsub("[^a-zA-z0-9]", "", type$name)
  folder <- file.path(cachedir, safename)
  created <- dir.create(folder, showWarnings = FALSE, recursive = TRUE)
  folder
}

tile.plotarray <- function(image, box) {
  graphics::rasterImage(image, box[1, 1], box[2, 1], box[1, 2], box[2, 2])
}

tile.autozoom <- function(res = 150, epsg = 4326) {
  ext <- graphics::par("usr")
  midy <- mean(c(ext[3], ext[4]))
  rightmid <- .tolatlon(ext[2], midy, epsg)
  centermid <- .tolatlon(mean(c(ext[1], ext[2])), midy, epsg)
  leftmid <- .tolatlon(ext[1], midy, epsg)

  anglewidth1 <- rightmid[1] - centermid[1]
  if (anglewidth1 < 0) {
    anglewidth1 <- anglewidth1 + 360
  }

  anglewidth2 <- rightmid[1] - centermid[1]
  if (anglewidth2 < 0) {
    anglewidth2 <- anglewidth2 + 360
  }
  anglewidth <- anglewidth1 + anglewidth2

  # PROBLEMS WITH WIDE EXTENTS LIKE THE WORLD
  widthin <- graphics::grconvertX(ext[2], from = "user", to = "inches") -
    graphics::grconvertX(ext[1], from = "user", to = "inches")
  widthpx <- widthin * res

  zoom <- log2((360.0 / anglewidth) * (widthpx / 256.0))

  as.integer(floor(zoom))
}

# function to extract the bounding box given an input object
# always returns the bbox in lat/lon

#' @rdname deprecated
#' @export
extract_bbox <- function(x, tolatlon = TRUE, ...) {
  if (methods::is(x, "Spatial")) {
    if (tolatlon && !is.na(x@proj4string@projargs)) {
      x <- sf::as_Spatial(sf::st_transform(sf::st_as_sfc(x), 4326))
    }

    sp::bbox(x)
  } else if (methods::is(x, "Raster")) {
    box <- raster::as.matrix(x@extent)
    if (tolatlon && !is.na(x@crs@projargs)) {
      # need a couple of points to get a decent approximation
      coords <- expand.grid(x = box[1, ], y = box[2, ])
      coords_lonlat <- .tolatlon(coords[, 1], coords[, 2], projection = x@crs)
      lon_range <- range(coords_lonlat[, 1])
      lat_range <- range(coords_lonlat[, 2])
      box <- makebbox(lat_range[2], lon_range[2], lat_range[1], lon_range[1])
    }

    box
  } else if (methods::is(x, "bbox")) {
    makebbox(x[4], x[3], x[2], x[1])
  } else if (is.character(x)) {
    stop("Lookup using prettymapr::searchbbox() is no longer supported")
  } else if (is.bbox(x)) {
    x
  } else {
    extract_bbox(sf::st_bbox(x))
  }
}

is.bbox <- function(x) {
  is.matrix(x) && identical(dim(x), c(2L, 2L)) && identical(rownames(x), c("x", "y"))
}

# function to extract a projection from an object
extract_projection <- function(x) {
  if (methods::is(x, "CRS")) {
    sf::st_crs(x)
  } else if (methods::is(x, "Spatial")) {
    if (!is.na(x@proj4string@projargs)) {
      x@proj4string
    } else {
      NA
    }
  } else if (methods::is(x, "Raster")) {
    sf::st_crs(x@crs)
  } else if (is.numeric(x) && (length(x) == 1)) {
    intx <- as.integer(x)
    sf::st_crs(intx)
  } else {
    NA
  }
}
# plot slippy tiles

tile.loadimage <- function(x, y, zoom, type, cachedir = NULL, quiet = TRUE) {
  if (x < 0) {
    # negative tiles from wrap situation
    x <- x + 2^zoom
  }

  fname <- tile.cachename(x, y, zoom, type, cachedir)
  parts <- strsplit(fname, "\\.")[[1]]
  ext <- parts[length(parts)]
  tryCatch(
    {
      if (ext == "jpg" || ext == "jpeg") {
        jpeg::readJPEG(fname)
      } else if (ext == "png") {
        png::readPNG(fname)
      } else {
        stop("Extension not recognized: ", ext)
      }
    },
    error = function(err) {
      if (!quiet) message("Error loading ", fname, ": ", err)
      NULL
    }
  )
}

tile.applywrap <- function(tiles, zoom) {
  if (!all(min(tiles[, 1]):max(tiles[, 1]) %in% tiles[, 1])) {
    # wrapping around the backside of the earth, make end tiles negative
    warning("Attempting to plot wrap around tiles (~lat=-180), things may get funky.")
    n <- -1
    while (length(tiles[, 1][tiles[, 1] == 2^zoom + n]) > 0) {
      tiles[, 1][tiles[, 1] == 2^zoom + n] <- (tiles[, 1][tiles[, 1] == 2^zoom + n]) - 2^zoom
      n <- n - 1
    }
  }
  tiles
}

# loops through the tiles and applies a function (returning a list)
tile.apply <- function(tiles, fun, ..., progress = "none") {
  plyr::alply(tiles, 1, function(tile, ...) {
    fun(tile[[1]], tile[[2]], ...)
  }, ..., .progress = progress)
}

# loops through the tiles and plots or combines the results to a list
tile.ploteach <- function(tiles, zoom, type, epsg = 4326, cachedir = NULL, quiet = FALSE) {
  tile.apply(tiles, function(x, y) {
    box <- tile.bbox(x, y, zoom, epsg)
    image <- tile.loadimage(x, y, zoom, type, cachedir, quiet = quiet)

    # if in plotting mode, plot the array
    if (!is.null(image)) tile.plotarray(image, box)
  })
}

tile.each <- function(tiles, zoom, type, epsg = 4326, cachedir = NULL, quiet = FALSE) {
  tile.apply(tiles, function(x, y) {
    box <- tile.bbox(x, y, zoom, epsg)
    image <- tile.loadimage(x, y, zoom, type, cachedir, quiet = quiet)

    # return structure as the image array, with attribute 'bbox'
    # this is modeled after the @bbox slot in the sp package
    structure(image,
      bbox = box, type = type,
      epsg = epsg, zoom = zoom, tiles = data.frame(x, y)
    )
  })
}

tile.fuse <- function(tiles, zoom, type, epsg = 4326, cachedir = NULL, quiet = FALSE) {
  tiles <- tile.applywrap(tiles, zoom)

  tile_dims <- check.dimensions(tiles, zoom, type, cachedir)
  dim_x <- tile_dims$targetdim[1]
  dim_y <- tile_dims$targetdim[2]
  dim_bands <- tile_dims$targetdim[3]

  if (tile_dims$nmissing > 0) {
    message(tile_dims$nmissing, " could not be loaded for type ", type$name)
    missing_tile <- array(0, tile_dims$targetdim)
  } else {
    missing_tile <- NULL
  }

  tiles <- tiles[order(tiles[[1]], tiles[[2]]), , drop = FALSE]
  x_tile <- unique(tiles[, 1])
  y_tile <- unique(tiles[, 2])

  xs <- (seq_along(x_tile) - 1) * dim_x
  ys <- (seq_along(y_tile) - 1) * dim_y

  out <- array(1, dim = c(dim_y * length(y_tile), dim_x * length(x_tile), dim_bands))

  for (i in seq_along(x_tile)) {
    for (j in seq_along(y_tile)) {
      img <- tile.loadimage(x_tile[i], y_tile[j], zoom, type, cachedir, quiet = quiet)
      if (is.null(img) && is.null(missing_tile)) {
        stop("Cannot fuse unloadable tile")
      } else if (is.null(img)) {
        missing_tile
      } else {
        ensure.bands(img, tile_dims$targetdim, default_value = 1)
      }

      # row, column, band (where row == y dimension)
      # sometimes both 3- and 4- band images are returned, so the assignment
      # needs to assign only the bands in the loaded img
      out[
        ys[j]:(ys[j] + dim_y - 1) + 1,
        xs[i]:(xs[i] + dim_x - 1) + 1,
        seq_len(dim(img)[3])
      ] <- img
    }
  }

  # calc bounding box of whole image
  nw <- tile.nw(min(x_tile), min(y_tile), zoom, epsg)
  se <- tile.nw(max(x_tile) + 1, max(y_tile) + 1, zoom, epsg)

  bbox <- matrix(
    c(nw[1], se[2], se[1], nw[2]),
    ncol = 2,
    byrow = FALSE,
    dimnames = list(
      c("x", "y"),
      c("min", "max")
    )
  )

  # return same structure as tile.each()
  structure(out, bbox = bbox, epsg = epsg, type = type, zoom = zoom, tiles = tiles)
}

# ensure array dimensions match a given dim value
ensure.bands <- function(image, dimension, default_value = 1) {
  banddiff <- dimension[3] - dim(image)[3]
  if (banddiff == 0) {
    image
  } else if (banddiff > 0) {
    # add extra bands
    abind::abind(image, array(
      default_value,
      c(dimension[1], dimension[2], banddiff)
    ),
    along = 3
    )
  } else if (banddiff < 0) {
    # this shouldn't happen, but...
    warning("Cropping image in ensure.bands")
    # crop
    image[, , 1:dimension[3], drop = FALSE]
  }
}


# checks the dimensions of all the tiles (used in tile.fuse)
check.dimensions <- function(tiles, zoom, type, cachedir) {
  # check dimensions of all tiles before fusing
  dims <- tile.apply(tiles, fun = function(x, y) {
    image <- tile.loadimage(x, y, zoom, type, cachedir)
    if (!is.null(image)) {
      dim(image)
    } else {
      c(0, 0, 0)
    }
  })

  # check for 3 dimensions
  if (!all(vapply(dims, length, integer(1)) == 3)) stop("Incorrect dimensions in image")

  # check for missing tiles
  missing_tiles <- vapply(
    dims, function(dim) identical(dim, c(0, 0, 0)),
    logical(1)
  )
  if (all(missing_tiles)) stop("Zero tiles were loaded for type ", type$name)

  # find dimension of non-missing tiles (hopefully the same...)
  tiledim <- do.call(rbind, dims[!missing_tiles])

  uniqueXs <- unique(tiledim[, 1, drop = TRUE])
  uniqueYs <- unique(tiledim[, 2, drop = TRUE])
  if (length(uniqueXs) > 1) {
    stop(
      "More than one image x dimension: ",
      paste(uniqueXs, collapse = ", ")
    )
  }
  if (length(uniqueYs) > 1) {
    stop(
      "More than one image y dimension: ",
      paste(uniqueYs, collapse = ", ")
    )
  }

  # assign target dim with the max of z dimensions (so a band can be added)
  # if not all have the same bands
  targetdim <- c(uniqueXs, uniqueYs, max(tiledim[, 3, drop = TRUE]))

  # also return the number of missing tiles
  list(targetdim = targetdim, nmissing = sum(missing_tiles))
}

tile.plotfused <- function(tiles, zoom, type, epsg = 4326, cachedir = NULL, quiet = FALSE) {
  fused <- tile.fuse(tiles, zoom, type, epsg = epsg, cachedir = cachedir, quiet = quiet)
  # plot image
  tile.plotarray(fused, attr(fused, "bbox"))
}

#' @rdname deprecated
#' @export
osm.plot <- function(bbox, zoomin = 0, zoom = NULL, type = NULL, forcedownload = FALSE,
                     stoponlargerequest = TRUE, fusetiles = TRUE, cachedir = NULL, res = 150,
                     project = TRUE, progress = c("text", "none"), quiet = TRUE,
                     api_key = NULL, ...) {
  # validate progress arg
  progress <- match.arg(progress)

  # get lookup information from input
  bbox <- extract_bbox(bbox)

  # verify tile source
  if (is.null(type)) {
    type <- get_default_tile_source()
  } else {
    type <- as.tile_source(type, api_key = api_key)
  }

  if (project) {
    epsg <- 3857
  } else {
    epsg <- 4326
  }

  bbox <- .projectbbox(bbox, epsg)

  coords <- sp::coordinates(t(bbox))
  spoints <- sp::SpatialPoints(coords)

  plotargs <- list(...)
  if (is.null(plotargs$xlim)) {
    xlim <- bbox[1, ]
  }
  if (is.null(plotargs$ylim)) {
    ylim <- bbox[2, ]
  }

  sp::plot(spoints, pch = ".", xlim = xlim, ylim = ylim, ...)

  if (is.null(zoom)) {
    zoom <- tile.autozoom(res = res, epsg = epsg)
  }
  zoom <- zoom + zoomin
  maxzoom <- tile.maxzoom(type)
  zoom <- min(zoom, maxzoom)

  # global min zoom set to 1
  zoom <- max(1, zoom)
  message("Zoom: ", zoom)
  # adjust bbox to final plot extents
  bbox <- t(matrix(graphics::par("usr"), ncol = 2, byrow = FALSE))

  tiles <- tiles.bybbox(bbox, zoom, epsg = epsg)
  if ((nrow(tiles) > 32) && stoponlargerequest) {
    stop(
      "More than 32 tiles to be loaded. ",
      "Run with stoponlargerequest=FALSE or ",
      "zoomin=-1, to continue"
    )
  }
  tile.download(tiles, zoom,
    type = type, forcedownload = forcedownload, cachedir = cachedir,
    progress = progress, quiet = quiet
  )

  if (fusetiles) {
    tile.plotfused(tiles, zoom, type = type, epsg = epsg, cachedir = cachedir, quiet = quiet)
  } else {
    tile.ploteach(tiles, zoom, type = type, epsg = epsg, cachedir = cachedir, quiet = quiet)
  }

  tile.attribute(type)
}


tile.raster.autozoom <- function(bbox, epsg, minnumtiles = 12) {
  for (zoom in 1:19) {
    tiles <- tiles.bybbox(bbox, zoom, epsg) # always latlon
    if (nrow(tiles) >= minnumtiles) {
      return(zoom)
    }
  }
  return(20)
}


#' @rdname deprecated
#' @export
osm.image <- function(x, zoomin = 0, zoom = NULL, type = NULL, forcedownload = FALSE, cachedir = NULL,
                      progress = c("text", "none"), quiet = TRUE, api_key = NULL) {
  # verify progress input
  progress <- match.arg(progress)

  # verify tile source
  if (is.null(type)) {
    type <- get_default_tile_source()
  } else {
    type <- as.tile_source(type, api_key = api_key)
  }

  # get lookup information from input
  lookup.bbox <- extract_bbox(x)

  # get zoom level
  if (is.null(zoom)) {
    zoom <- tile.raster.autozoom(lookup.bbox, epsg = 4326)
  }

  zoom <- min(zoom + zoomin, tile.maxzoom(type))
  zoom <- max(1, zoom) # global min zoom set to 1
  if (progress != "none") message("Zoom: ", zoom)

  # get tile list, download tiles
  tiles <- tiles.bybbox(lookup.bbox, zoom, epsg = 4326)
  tile.download(tiles, zoom,
    type = type, forcedownload = forcedownload, cachedir = cachedir,
    progress = progress, quiet = quiet
  )

  # return fused image
  tile.fuse(tiles, zoom, type = type, epsg = 3857, cachedir = cachedir, quiet = quiet)
}

#' @rdname deprecated
#' @export
osm.raster <- function(x, zoomin = 0, zoom = NULL, type = "osm", forcedownload = FALSE, cachedir = NULL,
                       progress = c("text", "none"), quiet = TRUE,
                       projection = NULL, crop = FALSE, filename = NULL, resample = "bilinear",
                       api_key = NULL, ...) {
  # verify inputs
  if (!requireNamespace("raster")) stop("Package 'raster' is required for osm.raster()")

  # verify progress input
  progress <- match.arg(progress)

  # verify tile source
  if (is.null(type)) {
    type <- get_default_tile_source()
  } else {
    type <- as.tile_source(type, api_key = api_key)
  }

  # if filename is specified, write output of osm.raster() to filename
  if (!is.null(filename)) {
    if (methods::is(x, "Raster")) {
      return(invisible(raster::writeRaster(x, filename = filename, datatype = "INT1U", ...)))
    } else {
      return(invisible(raster::writeRaster(
        osm.raster(
          x = x, zoomin = zoomin, zoom = zoom,
          type = type, forcedownload = forcedownload,
          progress = progress, quiet = quiet,
          cachedir = cachedir, projection = projection,
          crop = crop, filename = NULL, api_key = api_key
        ),
        filename = filename, datatype = "INT1U", ...
      )))
    }
  }

  # extract source projection
  src_proj <- extract_projection(x)

  # extract bounding box
  lookup.bbox <- extract_bbox(x)
  # if input was a character, make sure lookup only happens once
  if (is.character(x)) {
    x <- lookup.bbox
  }

  # fill in destination projection if not specified
  if (is.null(projection)) {
    if (is.na(src_proj)) {
      # default is to project to google mercator
      projection <- extract_projection(3857)
    } else {
      projection <- src_proj
    }
  } else {
    projection <- extract_projection(projection)
  }

  # get cropping information, if desired
  if (crop) {
    crop.bbox <- extract_bbox(x, tolatlon = FALSE)
    if (is.na(src_proj)) {
      # default is to project to google mercator
      crop.bbox <- .projectbbox(crop.bbox, projection = projection)
    }
  }

  arr <- osm.image(x,
    zoomin = zoomin, zoom = zoom, type = type,
    forcedownload = forcedownload, cachedir = cachedir,
    progress = progress, quiet = quiet
  )
  box <- attr(arr, "bbox")
  nbrow <- dim(arr)[1]
  nbcol <- dim(arr)[2]
  bands <- dim(arr)[3]
  .makeraster <- function(i) {
    raster::raster(matrix(arr[, , i] * 255, nrow = nbrow, ncol = nbcol),
      xmn = box[1, 1], xmx = box[1, 2],
      ymn = box[2, 1], ymx = box[2, 2],
      crs = "EPSG:3857"
    )
  }
  if (bands == 3) {
    rstack <- raster::stack(.makeraster(1), .makeraster(2), .makeraster(3))
  } else if (bands == 4) {
    rstack <- raster::stack(.makeraster(1), .makeraster(2), .makeraster(3), .makeraster(4))
  } else {
    stop("Invalid number of bands in image: ", bands)
  }

  if (!is.null(projection)) {
    if (crop) {
      return(osm.proj(rstack, projection, crop.bbox, method = resample))
    } else {
      return(osm.proj(rstack, projection, method = resample))
    }
  } else {
    if (crop) {
      return(osm.proj(rstack, "EPSG:3857", crop.bbox, method = resample))
    } else {
      return(rstack)
    }
  }
}

# @title Project an OSM RasterStack
# projects a raster stack generated above
osm.proj <- function(osm.raster, projection, crop.bbox = NULL, ...) {
  projection <- sf::st_crs(projection)
  if (sf::st_crs(osm.raster@crs) == projection) {
    rstackproj <- osm.raster
  } else {
    osm.terra <- terra::rast(osm.raster)
    terra::crs(osm.terra) <- terra::crs("EPSG:3857")
    rstackproj <- raster::stack(
      terra::project(
        osm.terra,
        terra::crs(sf::st_crs(projection)$wkt),
        ...
      )
    )
  }

  # this can occur because of bilinear resampling on project
  # values outside [0, 255] cause problems (e.g., raster::plotRGB())
  rstackproj <- raster::clamp(rstackproj, lower = 0, upper = 255)

  if (!is.null(crop.bbox)) {
    k <- min(c(
      0.052 * (crop.bbox[2, 2] - crop.bbox[2, 1]),
      0.052 * (crop.bbox[1, 2] - crop.bbox[1, 1])
    ))
    crop.bbox[2, 2] <- crop.bbox[2, 2] + k
    crop.bbox[1, 1] <- crop.bbox[1, 1] - k
    crop.bbox[2, 1] <- crop.bbox[2, 1] - k
    crop.bbox[1, 2] <- crop.bbox[1, 2] + k

    return(raster::crop(rstackproj, crop.bbox))
  } else {
    return(rstackproj)
  }
}

#' @rdname deprecated
#' @export
osm.points <- function(x, y = NULL, epsg = 4326, toepsg = 3857, ...) {
  graphics::points(.projpts(x, y, epsg, toepsg), ...)
}

#' @rdname deprecated
#' @export
osm.segments <- function(x0, y0, x1 = x0, y1 = y0, epsg = 4326, toepsg = 3857, ...) {
  c0 <- .projpts(x0, y0, epsg, toepsg)
  c1 <- .projpts(x1, y1, epsg, toepsg)
  graphics::segments(c0[, 1], c0[, 2], c1[, 1], c1[, 2])
}

#' @rdname deprecated
#' @export
osm.lines <- function(x, y = NULL, epsg = 4326, toepsg = 3857, ...) {
  graphics::lines(.projpts(x, y, epsg, toepsg), ...)
}

#' @rdname deprecated
#' @export
osm.polygon <- function(x, y = NULL, epsg = 4326, toepsg = 3857, ...) {
  graphics::polygon(.projpts(x, y, epsg, toepsg), ...)
}


#' @rdname deprecated
#' @export
osm.text <- function(x, y = NULL, labels = seq_along(x), epsg = 4326, toepsg = 3857, ...) {
  graphics::text(.projpts(x, y, epsg, toepsg), labels = labels, ...)
}


.projpts <- function(x, y = NULL, epsg, toepsg) {
  xy <- grDevices::xy.coords(x, y)
  sf::sf_project(
    paste0("EPSG:", epsg),
    paste0("EPSG:", toepsg),
    cbind(xy$x, xy$y)
  )
}

# projection functions

.tolatlon <- function(x, y, epsg = NULL, projection = NULL) {
  if (is.null(epsg) && is.null(projection)) {
    stop("epsg and projection both null...nothing to project")
  } else if (!is.null(epsg) && !is.null(projection)) {
    stop("epsg and projection both specified...ambiguous call")
  }

  if (is.null(projection)) {
    projection <- sf::st_crs(paste0("EPSG:", epsg))
  } else {
    projection <- sf::st_crs(projection)
  }

  sf::sf_project(
    projection,
    sf::st_crs(4326),
    cbind(x, y)
  )
}

.fromlatlon <- function(lon, lat, epsg = NULL, projection = NULL) {
  if (is.null(epsg) && is.null(projection)) {
    stop("epsg and projection both null...nothing to project")
  } else if (!is.null(epsg) && !is.null(projection)) {
    stop("epsg and projection both specified...ambiguous call")
  }

  if (is.null(projection)) {
    projection <- sf::st_crs(epsg)
  } else {
    projection <- sf::st_crs(projection)
  }

  sf::sf_project(
    sf::st_crs(4326),
    projection,
    cbind(lon, lat)
  )
}

.projectbbox <- function(bbox, toepsg = NULL, projection = NULL) {
  if (is.null(toepsg) && is.null(projection)) {
    stop("toepsg and projection both null...nothing to project")
  } else if (!is.null(toepsg) && !is.null(projection)) {
    stop("toepsg and projection both specified...ambiguous call")
  }

  if (is.null(projection)) {
    projection <- sf::st_crs(toepsg)
  } else {
    projection <- sf::st_crs(projection)
  }

  newpoints <- sf::sf_project(
    sf::st_crs(4326),
    projection,
    t(bbox)
  )

  newbbox <- t(newpoints)

  if (newbbox[1, 1] > newbbox[1, 2]) { # if min>max
    maxx <- .fromlatlon(180, bbox[2, 1], projection = projection)[1]
    newbbox[1, 1] <- newbbox[1, 1] - maxx * 2
  }
  newbbox
}

.revprojectbbox <- function(bbox, fromepsg = NULL, projection = NULL) {
  if (is.null(fromepsg) && is.null(projection)) {
    stop("fromepsg and projection both null...nothing to project")
  } else if (!is.null(fromepsg) && !is.null(projection)) {
    stop("fromepsg and projection both specified...ambiguous call")
  }

  if (is.null(projection)) {
    projection <- sf::st_crs(fromepsg)
  } else {
    projection <- sf::st_crs(projection)
  }

  newpoints <- sf::sf_project(
    projection,
    sf::st_crs(4326),
    t(bbox)
  )

  t(newpoints)
}

# tile URLs
# create list() of tile sources
tile_sources <- list(
  osm = source_from_url_format(
    url_format = c(
      "http://a.tile.openstreetmap.org/${z}/${x}/${y}.png",
      "http://b.tile.openstreetmap.org/${z}/${x}/${y}.png",
      "http://c.tile.openstreetmap.org/${z}/${x}/${y}.png"
    ),
    attribution = paste(
      "Consider donating to the Open Street Map project",
      "at http://donate.openstreetmap.org/"
    )
  ),
  opencycle = source_from_url_format(
    url_format = c(
      "http://a.tile.opencyclemap.org/cycle/${z}/${x}/${y}.png",
      "http://b.tile.opencyclemap.org/cycle/${z}/${x}/${y}.png"
    ),
    attribution = paste(
      "Consider donating to the Open Street Map project",
      "at http://donate.openstreetmap.org/"
    )
  ),
  hotstyle = source_from_url_format(
    url_format = c(
      "http://a.tile.openstreetmap.fr/hot/${z}/${x}/${y}.png",
      "http://b.tile.openstreetmap.fr/hot/${z}/${x}/${y}.png"
    ),
    attribution = paste(
      "Consider donating to the Open Street Map project",
      "at http://donate.openstreetmap.org/"
    )
  ),
  loviniahike = source_from_url_format(
    url_format = "http://tile.waymarkedtrails.org/hiking/${z}/${x}/${y}.png",
    attribution = "Visit the Lovinia hike map at http://osm.lonvia.de/hiking.html"
  ),
  loviniacycle = source_from_url_format(
    url_format = "http://tile.waymarkedtrails.org/cycling/${z}/${x}/${y}.png",
    attribution = "Visit the Lovinia cycle map at http://cycling.waymarkedtrails.org/"
  ),
  stamenbw = source_from_url_format(
    url_format = "http://a.tile.stamen.com/toner/${z}/${x}/${y}.png",
    attribution = "Visit the Stamen cartography website at http://maps.stamen.com/"
  ),
  stamenwatercolor = source_from_url_format(
    url_format = "http://a.tile.stamen.com/watercolor/${z}/${x}/${y}.jpg",
    attribution = "Visit the Stamen cartography website at http://maps.stamen.com/"
  ),
  osmtransport = source_from_url_format(
    url_format = c(
      "http://a.tile2.opencyclemap.org/transport/${z}/${x}/${y}.png",
      "http://b.tile2.opencyclemap.org/transport/${z}/${x}/${y}.png"
    ),
    attribution = paste(
      "Consider donating to the Open Street Map project",
      "at http://donate.openstreetmap.org/"
    )
  ),
  thunderforestlandscape = source_from_url_format(
    url_format = c(
      "http://a.tile.thunderforest.com/landscape/${z}/${x}/${y}.png",
      "http://b.tile.thunderforest.com/landscape/${z}/${x}/${y}.png",
      "http://c.tile.thunderforest.com/landscape/${z}/${x}/${y}.png"
    ),
    attribution = "More on Thunderforest at http://www.thunderforest.com/"
  ),
  thunderforestoutdoors = source_from_url_format(
    url_format = c(
      "http://a.tile.thunderforest.com/outdoors/${z}/${x}/${y}.png",
      "http://b.tile.thunderforest.com/outdoors/${z}/${x}/${y}.png",
      "http://c.tile.thunderforest.com/outdoors/${z}/${x}/${y}.png"
    ),
    attribution = "More on Thunderforest at http://www.thunderforest.com/"
  ),
  cartodark = source_from_url_format(
    url_format = "https://a.basemaps.cartocdn.com/dark_all/${z}/${x}/${y}.png",
    attribution = "Map tiles by Carto, under CC BY 3.0. Data by OpenStreetMap, under ODbL."
  ),
  cartolight = source_from_url_format(
    url_format = "https://a.basemaps.cartocdn.com/light_all/${z}/${x}/${y}.png",
    attribution = "Map tiles by Carto, under CC BY 3.0. Data by OpenStreetMap, under ODbL."
  )
)

#' @rdname deprecated
#' @export
makebbox <- function(n, e, s, w) {
  if (isTRUE(n < s)) {
    warning("North less than south. Check order?")
  }
  matrix(
    c(w, s, e, n),
    byrow = FALSE,
    ncol = 2,
    dimnames = list(c(
      "x",
      "y"
    ), c("min", "max"))
  )
}

#' @rdname deprecated
#' @export
zoombbox <- function(bbox, factor = 1, offset = c(0, 0)) {
  lons <- bbox[1, ]
  lats <- bbox[2, ]
  clon <- mean(lons) + offset[1]
  clat <- mean(lats) + offset[2]
  if (length(factor) > 1) {
    factorx <- factor[1]
    factory <- factor[2]
  } else {
    factorx <- factor
    factory <- factor
  }
  newwidth <- (lons[2] - lons[1]) / factorx
  newheight <- (lats[2] - lats[1]) / factory
  makebbox(
    clat + newheight / 2, clon + newwidth / 2, clat - newheight / 2,
    clon - newwidth / 2
  )
}

test_that("all named tile sources load", {
  skip_if_not(identical(Sys.getenv("R_ROSM_TEST_DEPRECATED"), "true"))
  nsbox <- makebbox(47.2, -59.7, 43.3, -66.4)

  # test contributed by rCarto (cartography package)

  unavailable <- c("stamenbw", "stamenwatercolor")
  if (!nzchar(Sys.getenv("CARTO_API_KEY"))) {
    unavailable <- c(unavailable, "cartodark", "cartolight")
  }

  tiles <- data.frame(
    # Stamen maps no longer load; CARTO requires a user-supplied API key.
    types = setdiff(osm.types(), unavailable),
    status = NA,
    stringsAsFactors = FALSE
  )

  # helpful to run this on travis builds, since they are not often
  expect_message({
    for (i in 1:nrow(tiles)) {
      status <- tryCatch(
        {
          osm.plot(nsbox, type = tiles[i, 1], xlab = tiles[i, 1], progress = "none")
          status <- "OK"
        },
        error = function(cond) {
          message(cond, "\n")
          status <- "error"
        }
      )
      tiles[i, "status"] <- status
    }
  })

  expect_true(all(tiles$status == "OK"))
})

test_that("all named bing tile sources load", {
  skip_if_not(identical(Sys.getenv("R_ROSM_TEST_DEPRECATED"), "true"))
  skip("As of 2026-01-20 all Bing map types fail to load")
  nsbox <- makebbox(47.2, -59.7, 43.3, -66.4)

  # test contributed by rCarto (cartography package)

  tiles <- data.frame(
    types = bmaps.types(), status = NA,
    stringsAsFactors = FALSE
  )

  expect_message({
    for (i in 1:nrow(tiles)) {
      status <- tryCatch(
        {
          bmaps.plot(nsbox, type = tiles[i, 1], progress = "none")
          status <- "OK"
        },
        error = function(cond) {
          message(cond, "\n")
          status <- "error"
        }
      )
      tiles[i, "status"] <- status
    }
  })

  expect_true(all(tiles$status == "OK"))
})

test_that("custom map types load tiles", {
  skip_if_not(identical(Sys.getenv("R_ROSM_TEST_DEPRECATED"), "true"))
  nsbox <- makebbox(47.2, -59.7, 43.3, -66.4)

  # this should throw a deprecation message but still work
  tile.url.darkmatter <- function(xtile, ytile, zoom) {
    paste0(paste("http://a.basemaps.cartocdn.com/dark_all",
      zoom, xtile, ytile,
      sep = "/"
    ), ".png")
  }

  assign("tile.url.darkmatter", tile.url.darkmatter, .GlobalEnv)

  expect_message(
    osm.plot(nsbox, type = "darkmatter", progress = "none"),
    "Using functions as custom tile sources is deprecated. Use string formats instead."
  )

  rm("tile.url.darkmatter", envir = .GlobalEnv)

  # using string formats
  ts <- as.tile_source("http://a.basemaps.cartocdn.com/dark_all/${z}/${x}/${y}.png")
  expect_s3_class(ts, "tile_source")

  expect_equal(
    ts$get_tile_url(2, 1, 3),
    "http://a.basemaps.cartocdn.com/dark_all/3/2/1.png"
  )

  expect_message(osm.plot(nsbox, type = ts, progress = "none"))

  # test registration
  register_tile_source(dark = "http://a.basemaps.cartocdn.com/dark_all/${z}/${x}/${y}.png")
  expect_true("dark" %in% names(rosm:::registered_sources))
  expect_is(rosm:::registered_sources$dark, "tile_source")

  expect_message(osm.plot(nsbox, type = "dark", progress = "none"))

  # test setting of default
  set_default_tile_source("dark")
  expect_equal(get_default_tile_source()$name, "dark")
  # plot
  expect_message(osm.plot(nsbox, progress = "none"))
  # reset tile source
  set_default_tile_source("osm")
})

test_that("CARTO map types require and use an API key", {
  withr::local_envvar(CARTO_API_KEY = NA)

  expect_error(
    as.tile_source("cartolight"),
    "Supply `api_key` or set the CARTO_API_KEY environment variable",
    fixed = TRUE
  )

  light <- as.tile_source("cartolight", api_key = "explicit key")
  expect_identical(
    light$get_tile_url(2, 1, 3),
    "https://a.basemaps.cartocdn.com/light_all/3/2/1.png?key=explicit%20key"
  )

  withr::local_envvar(CARTO_API_KEY = "environment key")
  dark <- as.tile_source("cartodark")
  expect_identical(
    dark$get_tile_url(2, 1, 3),
    "https://a.basemaps.cartocdn.com/dark_all/3/2/1.png?key=environment%20key"
  )
})


# plot the whole world (still doesn't work)
# osm.plot(makebbox(89.9, 179.9, -89.9, -179.9), zoom=0)
# prettymap(osm.plot(makebbox(89.9, 179.9, -89.9, -179.9)))

test_that("wrap around situations warn the user", {
  skip_if_not(identical(Sys.getenv("R_ROSM_TEST_DEPRECATED"), "true"))

  americas <- zoombbox(makebbox(89.9, 179.9, -89.9, -179.9), 2, c(-100, 0))
  # note this doesn't work with project = FALSE
  skip_if_not(identical(Sys.getenv("R_ROSM_TEST_DEPRECATED"), "true"))
  expect_warning(osm.plot(americas, zoomin = 1, progress = "none"), "Attempting to plot wrap around tiles")
  alaska <- makebbox(71, -129, 51, 172)
  expect_warning(osm.plot(alaska, progress = "none"), "Attempting to plot wrap around tiles")
  expect_warning(
    osm.plot(alaska, project = FALSE, progress = "none"),
    "Attempting to plot wrap around tiles"
  )
})

test_that("osm.image returns an image with the required attrs", {
  skip_if_not(identical(Sys.getenv("R_ROSM_TEST_DEPRECATED"), "true"))
  nsbox <- makebbox(47.2, -59.7, 43.3, -66.4)

  img <- osm.image(nsbox, progress = "none")
  expect_true(inherits(img, "array"))
  expect_length(dim(img), 3)
  expect_true(dim(img)[3] %in% c(3, 4))

  attrs <- attributes(img)
  expect_true(all(c("bbox", "epsg", "type", "zoom", "tiles") %in% names(attrs)))
  expect_true(inherits(attrs$bbox, "matrix"))
  expect_s3_class(attrs$type, "tile_source")
  expect_type(attrs$epsg, "double")
  expect_type(attrs$zoom, "double")
})

test_that("osm.raster creates raster objects with the correct projection", {
  skip_if_not(identical(Sys.getenv("R_ROSM_TEST_DEPRECATED"), "true"))
  nsbox <- makebbox(47.2, -59.7, 43.3, -66.4)

  # make more rasters (projected)
  x <- osm.raster(nsbox, projection = 32620, progress = "none", crop = TRUE)
  expect_s4_class(x, "Raster")
  expect_true(sf::st_crs(x) == sf::st_crs(32620))

  # expect_silent(raster::plotRGB(x))

  # make non-projected raster
  x <- osm.raster(nsbox, progress = "none")
  expect_s4_class(x, "Raster")
  expect_true(sf::st_crs(x) == sf::st_crs(3857))

  # expect_null(raster::plotRGB(x))

  # make cropped raster, ensure bounds are different
  x <- osm.raster(nsbox, progress = "none")
  xcropped <- osm.raster(nsbox, crop = TRUE, progress = "none")
  expect_false({
    boundsx <- raster::as.matrix(x@extent)
    boundscropped <- raster::as.matrix(xcropped@extent)
    all(boundsx == boundscropped)
  })
})

test_that("osm rasters are written to disk", {
  skip_if_not(identical(Sys.getenv("R_ROSM_TEST_DEPRECATED"), "true"))
  nsbox <- makebbox(47.2, -59.7, 43.3, -66.4)

  test_file <- tempfile(fileext = ".tif")
  x <- osm.raster(nsbox, progress = "none")
  osm.raster(x, filename = test_file, progress = "none", overwrite = TRUE)
  expect_true(file.exists(test_file))
  unlink(test_file)

  osm.raster(
    nsbox,
    projection = sp::CRS("+init=epsg:32620"), crop = TRUE,
    filename = test_file, overwrite = TRUE, progress = "none"
  )
  expect_true(file.exists(test_file))
  unlink(test_file)
})

# add test for default cachedir
test_that("non-default cache directories are respected", {
  skip_if_not(identical(Sys.getenv("R_ROSM_TEST_DEPRECATED"), "true"))
  nsbox <- makebbox(47.2, -59.7, 43.3, -66.4)

  # see issue #3 and PR #4
  # start with clean cache

  default_cache <- "rosm.cache"
  if (dir.exists(default_cache)) unlink(default_cache, recursive = TRUE)

  # osm.plot with default cache
  expect_false(dir.exists(default_cache))
  expect_message(osm.plot(nsbox, progress = "none"))
  expect_true(dir.exists(default_cache))
  unlink(default_cache, recursive = TRUE)

  # osm.plot with fusetiles = FALSE, default cache
  expect_false(dir.exists(default_cache))
  expect_message(osm.plot(nsbox, fusetiles = FALSE, progress = "none"))
  expect_true(dir.exists(default_cache))
  unlink(default_cache, recursive = TRUE)

  # osm.image with default cache
  expect_false(dir.exists(default_cache))
  osm.image(nsbox, progress = "none")
  expect_true(dir.exists(default_cache))
  unlink(default_cache, recursive = TRUE)

  # osm.plot with tempdir cache
  newcache <- tempfile()[1]
  expect_false(dir.exists(newcache))
  expect_message(osm.plot(nsbox, cachedir = newcache, progress = "none"))
  # expect no default cache
  expect_false(dir.exists(default_cache))
  # expect temporary cache
  expect_true(dir.exists(newcache))
  # remove tempfile cache
  unlink(newcache, recursive = TRUE)

  # osm.plot(..fusetiles = FALSE) with tempdir cache
  newcache <- tempfile()[1]
  expect_false(dir.exists(newcache))
  expect_message(osm.plot(nsbox, cachedir = newcache, fusetiles = FALSE, progress = "none"))
  # expect no default cache
  expect_false(dir.exists(default_cache))
  # expect temporary cache
  expect_true(dir.exists(newcache))
  # remove tempfile cache
  unlink(newcache, recursive = TRUE)

  # osm.image with tempdir cache
  newcache <- tempfile()[1]
  expect_false(dir.exists(newcache))
  osm.image(nsbox, cachedir = newcache, progress = "none")
  # expect no default cache
  expect_false(dir.exists(default_cache))
  # expect temporary cache
  expect_true(dir.exists(newcache))
  # remove tempfile cache
  unlink(newcache, recursive = TRUE)
})

test_that("default cache directory is respected", {
  skip_if_not(identical(Sys.getenv("R_ROSM_TEST_DEPRECATED"), "true"))
  nsbox <- makebbox(47.2, -59.7, 43.3, -66.4)

  default_cache <- get_default_cachedir()
  # check that default cache is the correct default
  expect_equal(default_cache, "rosm.cache")
  # start with a clean cache
  if (dir.exists(default_cache)) unlink(default_cache, recursive = TRUE)

  # osm.plot with default cache
  expect_false(dir.exists(default_cache))
  expect_message(osm.plot(nsbox, progress = "none"))
  expect_true(dir.exists(default_cache))
  unlink(default_cache, recursive = TRUE)

  # osm.plot with different default cache
  new_cache <- tempfile()[1]
  set_default_cachedir(new_cache)
  expect_identical(get_default_cachedir(), new_cache)

  expect_false(dir.exists(new_cache))
  expect_message(osm.plot(nsbox, progress = "none"))
  expect_true(dir.exists(new_cache))
  unlink(new_cache, recursive = TRUE)

  # check reset of default cache
  set_default_cachedir(NULL)
  expect_equal(default_cache, "rosm.cache")
})

test_that("extract bbox", {
  skip_if_not(identical(Sys.getenv("R_ROSM_TEST_DEPRECATED"), "true"))

  sf_bbox <- setNames(
    c(-89.60465, 43.03507, -89.20378, 43.36607),
    c("xmin", "ymin", "xmax", "ymax")
  )
  class(sf_bbox) <- "bbox"

  sp_bbox <- matrix(sf_bbox,
    ncol = 2,
    dimnames = list(c("x", "y"), c("min", "max"))
  )
  expect_equal(sp_bbox, extract_bbox(sf_bbox))
})
# some wrap around situations currently don't work
# test_that("harder wrap around situations do not fail", {
#   americas <- zoombbox(makebbox(89.9, 179.9, -89.9, -179.9), 2, c(-100, 0))
#   # this works but is mega off
#   expect_warning(osm.raster(americas), "Attempting to plot wrap around tiles")
#   x <- osm.raster(americas)
#   xbox <- raster::as.matrix(x@extent)
#   # should be square ish
#   wh <- xbox[, 2] - xbox[, 1]
#   aspect <- wh[1] /  wh[2]
#   expect_lte(abs(1-aspect), 0.2)
#
#   # this one doesn't work at all
#   expect_warning(osm.plot(americas, project = FALSE),
#                  "Attempting to plot wrap around tiles")
# })

#' Low-level tile math
#'
#' Open Street Map operates using a system of tiles whose value and bounds
#' are easily calculated from WGS84 longitude/latitude. These functions convert
#' between tile system coordinates and longitude/latitude.
#'
#' @param pt A vector of points as coerced by [wk::as_xy()]. The CRS for these
#'   points is considered.
#' @param zoom A zoom level, generally between 0 and 21, with higher values
#'   representing a smaller (i.e., more detailed) tile.
#' @param crs A target CRS. Either [wk::wk_crs_longlat()] or
#'   [osm_crs_native()].
#' @param tile A `data.frame()` with columns `x`, `y`, and `zoom`.
#'
#' @return
#'   - `osm_tile()`: A `data.frame()` with columns `x`, `y`, and `zoom`.
#'   - `osm_tile_top_left()`: A [wk::xy()] of the top-left (northwest) corner
#'     of the tile.
#'   - `osm_tile_envelope()`: A [wk::rct()] of the tile bounds.
#' @export
#'
#' @examples
#' (tiles <- osm_tile(osm_lnglat(-64, 45), zoom = 0:5))
#' osm_tile_envelope(tiles)
#'
osm_tile <- function(pt, zoom) {
  lnglat <- unclass(osm_ensure_lnglat(pt))

  lat_rad <- lnglat[[2]] * pi / 180.0
  n <- 2.0^zoom
  xtile <- floor(n / 2 + osm_lng_degrees_to_native(lnglat[[1]], n / pi / 2))
  ytile <- floor(n / 2 - osm_lat_degrees_to_native(lnglat[[2]], n / pi / 2))
  data.frame(x = xtile, y = ytile, zoom = zoom)
}

#' @rdname osm_tile
#' @export
osm_tile_quadkey <- function(tile) {
  ensure_tile(tile)

  n_tile <- nrow(tile)
  if (n_tile == 0) {
    return(character())
  }

  n <- 2^tile$zoom
  max_zoom <- max(tile$zoom)
  quadkey_blank <- strrep(" ", max_zoom)
  out <- rep(quadkey_blank, n_tile)

  keymap <- array(
    rep(c("0", "2", "1", "3"), n_tile),
    dim = c(2, 2, n_tile)
  )

  decx <- tile$x / n
  decy <- tile$y / n
  for (i in 1:max_zoom) {
    x <- floor(decx * 2^i) - floor(decx * 2^(i - 1)) * 2
    y <- floor(decy * 2^i) - floor(decy * 2^(i - 1)) * 2
    keymap_indices0 <- (seq_len(n_tile) - 1) * 4 + x * 2 + y
    substr(out, i, i) <- keymap[keymap_indices0 + 1]
  }

  substr(out, 1, tile$zoom)
}

#' @rdname osm_tile
#' @export
osm_tile_top_left <- function(tile, crs = osm_crs_native()) {
  ensure_tile(tile)
  n <- 2.0^tile$zoom

  if (crs_is_lnglat(crs)) {
    x <- osm_native_to_lng_degrees(tile$x - n / 2, n / pi / 2)
    y <- osm_native_to_lat_degrees(n / 2 - tile$y, n / pi / 2)
  } else if (crs_is_native(crs)) {
    x <- (tile$x / n - 0.5) * 2 * pi * 6378137
    y <- (0.5 - tile$y / n) * 2 * pi * 6378137
  } else {
    stop("Unsupported `crs`")
  }

  wk::xy(x, y, crs = crs)
}

#' @rdname osm_tile
#' @export
osm_tile_envelope <- function(tile, crs = osm_crs_native()) {
  ensure_tile(tile)
  top_left <- unclass(osm_tile_top_left(tile, crs = crs))
  bottom_right <- unclass(
    osm_tile_top_left(
      list(
        x = tile$x + 1,
        y = tile$y + 1,
        zoom = tile$zoom
      ),
      crs = crs
    )
  )

  wk::new_wk_rct(
    list(
      xmin = top_left$x,
      ymin = bottom_right$y,
      xmax = bottom_right$x,
      ymax = top_left$y
    ),
    crs = crs
  )
}

osm_tile_normalize <- function(tile) {
  ensure_tile(tile)
  n <- 2^(tile$zoom)

  # In the x direction, out-of-bounds tiles can be wrapped
  tile$x <- tile$x %% n

  # In the y direction, we just want these tiles to not exist
  tile$y[tile$y < 0] <- NA_real_
  tile$y[tile$y >= n] <- NA_real_

  tile
}

#' Get an OSM tile covering
#'
#' @param bbox A [wk::rct()] or object with a [wk::wk_bbox()] method.
#' @param zoom A zoom level or an auto zoom specifier like
#'   [osm_zoom_num_tiles()].
#' @param num_tiles The minimum number of tiles to use when choosing a
#'   zoom level.
#'
#' @return
#'   - `osm_tile_covering()` returns a `data.frame()` with columns x, y, and zoom.
#' @export
#'
#' @examples
#' bounds <- wk::rct(
#'   -7514064, 5009380,
#'   -6261722, 6261715,
#'   crs = osm_crs_native()
#' )
#'
#' osm_tile_covering(bounds)
#'
osm_tile_covering <- function(bbox, zoom = osm_zoom_num_tiles(6)) {
  # S2's s2_bounds_rect() gives a data.frame like this
  if (identical(names(bbox), c("lng_lo", "lat_lo", "lng_hi", "lat_hi"))) {
    names(bbox) <- c("xmin", "ymin", "xmax", "ymax")
    bbox <- wk::new_wk_rct(bbox, crs = wk::wk_crs_longlat())
  }

  crs <- wk::wk_crs(bbox)

  if (!inherits(bbox, "wk_rct")) {
    bbox <- wk::wk_bbox(bbox)
  }

  if (is.function(zoom)) {
    for (potential_zoom in 0:22) {
      potential_covering <- osm_tile_covering(bbox, potential_zoom)
      if (zoom(bbox, potential_covering)) {
        return(potential_covering)
      }
    }

    stop("`zoom` function returned FALSE for all zoom levels in 0:22")
  }

  if (crs_is_lnglat(crs)) {
    bbox <- unclass(bbox)
    # for these bboxes, xmin can be greater than xmax to wrap around a date line
    if (bbox$xmin > bbox$xmax) {
      bbox$xmin <- bbox$xmin - 180.0
    }

    # important to clamp latitude bounds or else we may get NaNs which prevent
    # us from getting a range of tile y values
    if (bbox$ymax > 85.0511287798065) {
      bbox$ymax <- 85.0511287798065
    }

    if (bbox$ymin < -85.0511287798065) {
      bbox$ymin <- -85.0511287798065
    }

    top_left <- osm_tile(wk::xy(bbox$xmin, bbox$ymax, crs = crs), zoom)
    bottom_right <- osm_tile(wk::xy(bbox$xmax, bbox$ymin, crs = crs), zoom)

    x <- seq(top_left$x, bottom_right$x)
    y <- seq(top_left$y, bottom_right$y)

    data.frame(
      x = rep(x, length(y)),
      y = rep(y, each = length(x)),
      zoom = zoom
    )
  } else if (crs_is_native(crs)) {
    bbox <- unclass(bbox)
    top_left <- osm_tile(wk::xy(bbox$xmin, bbox$ymax, crs = crs), zoom)
    bottom_right <- osm_tile(wk::xy(bbox$xmax, bbox$ymin, crs = crs), zoom)

    x <- seq(top_left$x, bottom_right$x)
    y <- seq(top_left$y, bottom_right$y)

    data.frame(
      x = rep(x, length(y)),
      y = rep(y, each = length(x)),
      zoom = zoom
    )
  } else {
    # Make a grid along the bounding box and compute the covering of that
    # after converting to native. This probably generates a covering that has
    # too many tiles in most cases.
    bbox_grd <- wk::grd(bbox, nx = 5, ny = 5, type = "corners")
    vertices_native <- osm_ensure_native(bbox_grd)
    osm_tile_covering(vertices_native, zoom = zoom)
  }
}

#' @rdname osm_tile_covering
#' @export
osm_zoom_num_tiles <- function(num_tiles) {
  force(num_tiles)
  function(bbox, result) {
    nrow(result) >= num_tiles
  }
}

#' Coordinate helpers
#'
#' @param x,y Ordinate values in EPSG:3857 (Spherical Mercator in meters)
#' @param lng,lat Coordinate values for longitude/latitude in degrees.
#' @inheritParams osm_tile
#'
#' @return
#'   - `osm_native()`, `osm_lnglat()`, `osm_ensure_native()`, and
#'     `osm_ensure_lnglat()` return a [wk::xy()] with the appropriate
#'     crs
#'   - `osm_crs_native()` returns a value that can be used as the [wk::wk_crs()]
#'     of a vector.
#' @export
#'
#' @examples
#' osm_lnglat(-64, 45)
#' osm_ensure_native(osm_lnglat(-64, 45))
#' osm_ensure_lnglat(
#'   osm_ensure_native(osm_lnglat(-64, 45))
#' )
osm_native <- function(x, y) {
  wk::xy(x, y, crs = osm_crs_native())
}

#' @rdname osm_native
#' @export
osm_lnglat <- function(lng, lat) {
  wk::xy(lng, lat, crs = wk::wk_crs_longlat())
}

#' @rdname osm_native
#' @export
osm_crs_native <- function() {
  "EPSG:3857"
}

#' @rdname osm_native
#' @export
osm_ensure_lnglat <- function(pt) {
  pt <- wk::as_xy(pt)
  crs <- wk::wk_crs(pt)
  if (is.null(crs)) {
    stop("Can't transform NULL crs to lon/lat")
  }

  if (inherits(crs, "wk_crs_inherit") || crs_is_lnglat(crs)) {
    wk::wk_set_crs(pt, wk::wk_crs_longlat())
  } else if (crs_is_native(crs)) {
    wk::xy(
      osm_native_to_lng_degrees(wk::xy_x(pt), scale = 6378137),
      osm_native_to_lat_degrees(wk::xy_y(pt), scale = 6378137),
      crs = wk::wk_crs_longlat()
    )
  } else {
    out <- sf::sf_project(
      wk::wk_crs_proj_definition(crs, verbose = TRUE),
      "EPSG:4326",
      as.matrix(pt),
      keep = TRUE,
      warn = FALSE,
      authority_compliant = FALSE
    )

    wk::new_wk_xy(
      list(
        x = out[, 1, drop = TRUE],
        y = out[, 2, drop = TRUE]
      ),
      crs = wk::wk_crs_longlat()
    )
  }
}

#' @rdname osm_native
#' @export
osm_ensure_native <- function(pt) {
  pt <- wk::as_xy(pt)
  crs <- wk::wk_crs(pt)
  if (is.null(crs)) {
    stop("Can't transform NULL crs to EPSG:3857")
  }

  if (inherits(crs, "wk_crs_inherit") || crs_is_native(crs)) {
    wk::wk_set_crs(pt, osm_crs_native())
  } else if (crs_is_lnglat(crs)) {
    wk::xy(
      osm_lng_degrees_to_native(wk::xy_x(pt), scale = 6378137),
      osm_lat_degrees_to_native(wk::xy_y(pt), scale = 6378137),
      crs = osm_crs_native()
    )
  } else {
    out <- sf::sf_project(
      wk::wk_crs_proj_definition(crs, verbose = TRUE),
      "EPSG:3857",
      as.matrix(pt),
      keep = TRUE,
      warn = FALSE,
      authority_compliant = FALSE
    )

    wk::new_wk_xy(
      list(
        x = out[, 1, drop = TRUE],
        y = out[, 2, drop = TRUE]
      ),
      crs = osm_crs_native()
    )
  }
}

crs_equal_any <- function(crs, dest_crs) {
  any(vapply(dest_crs, wk::wk_crs_equal, logical(1), crs))
}

crs_is_lnglat <- function(crs) {
  crs_equal_any(crs, list(wk::wk_crs_longlat(), "EPSG:4326"))
}

crs_is_native <- function(crs) {
  crs_equal_any(crs, list(osm_crs_native()))
}

ensure_tile <- function(tile) {
  if (!all(c("x", "y", "zoom") %in% names(tile))) {
    stop("`tile` must have columns `x`, `y`, and `zoom`")
  }

  if (!is.numeric(tile$x) || !is.numeric(tile$y) || !is.numeric(tile$zoom)) {
    stop("`tile$x`, `tile$y`, and `tile$zoom` must be numeric")
  }

  invisible(tile)
}

# Variants of projection functions whose projected range is [-pi -pi pi pi]
osm_lng_degrees_to_native <- function(lng_deg, scale = 1) {
  lng_deg / 180.0 * pi * scale
}

osm_lat_degrees_to_native <- function(lat_deg, scale = 1) {
  lat_rad <- lat_deg * pi / 180.0
  log(tan(lat_rad) + (1 / cos(lat_rad))) * scale
}

osm_native_to_lng_degrees <- function(x, scale = 1) {
  180.0 * x / pi / scale
}

osm_native_to_lat_degrees <- function(y, scale = 1) {
  atan(sinh(y / scale)) * 180 / pi
}

test_that("osm_tile() works", {
  expect_identical(
    osm_tile(osm_lnglat(-64, 45), 1:5),
    data.frame(
      x = c(0, 1, 2, 5, 10),
      y = c(0, 1, 2, 5, 11),
      zoom = 1:5
    )
  )
})

test_that("osm_tile_quadkey() works", {
  tiles <- data.frame(
    x = c(20, 21, 20, 21, 10, 0),
    y = c(22, 22, 23, 23, 11, 0),
    zoom = c(6, 6, 6, 6, 5, 0)
  )

  expect_identical(
    osm_tile_quadkey(tiles),
    c("030320", "030321", "030322", "030323", "03032", "")
  )

  expect_identical(
    osm_tile_quadkey(tiles[integer(0), , drop = FALSE]),
    character()
  )
})

test_that("osm_tile_covering() works for native bounds", {
  bounds <- wk::rct(
    -20037507, 1, -1, 20037507,
    crs = osm_crs_native()
  )

  expect_identical(
    osm_tile_covering(bounds, zoom = 1L),
    data.frame(
      x = 0L,
      y = 0L,
      zoom = 1L
    )
  )

  expect_identical(
    osm_tile_covering(bounds, zoom = 2L),
    data.frame(
      x = c(0L, 1L, 0L, 1L),
      y = c(0L, 0L, 1L, 1L),
      zoom = 2L
    )
  )
})

test_that("osm_tile_covering() works for lnglat bounds", {
  bounds <- wk::rct(
    -179.9, 0.1, -0.1, 85.05,
    crs = wk::wk_crs_longlat()
  )

  expect_identical(
    osm_tile_covering(bounds, zoom = 1L),
    data.frame(
      x = 0L,
      y = 0L,
      zoom = 1L
    )
  )

  expect_identical(
    osm_tile_covering(bounds, zoom = 2L),
    data.frame(
      x = c(0L, 1L, 0L, 1L),
      y = c(0L, 0L, 1L, 1L),
      zoom = 2L
    )
  )
})

test_that("osm_tile_covering() works for s2-style rectangle bounds", {
  bounds <- data.frame(
    lng_lo = 19,
    lat_lo = 41,
    lng_hi = -169,
    lat_hi = 81
  )

  bounds_rct <- wk::rct(19, 41, -169, 81, crs = wk::wk_crs_longlat())

  expect_identical(
    osm_tile_covering(bounds, zoom = 3L),
    osm_tile_covering(bounds_rct, zoom = 3L)
  )
})

test_that("osm_tile_covering() can compute a range for out of bounds lats", {
  bounds <- wk::rct(
    -179.9, -100, -0.1, 100,
    crs = wk::wk_crs_longlat()
  )

  expect_identical(
    osm_tile_covering(bounds, zoom = 1L),
    data.frame(
      x = c(0L, 0L),
      y = c(0L, 1L),
      zoom = 1L
    )
  )
})

test_that("osm_tile_covering() works for arbitrary CRS bounds", {
  bounds <- wk::rct(
    252185, 4815826, 739729, 5210280,
    crs = "EPSG:32620"
  )

  expect_identical(
    osm_tile_covering(bounds, zoom = 1L),
    data.frame(
      x = 0L,
      y = 0L,
      zoom = 1L
    )
  )

  expect_identical(
    osm_tile_covering(bounds, zoom = 6L),
    data.frame(
      x = c(20L, 21L, 20L, 21L),
      y = c(22L, 22L, 23L, 23L),
      zoom = 6L
    )
  )
})

test_that("osm_tile_covering() works with zoom as a function", {
  bounds <- wk::rct(
    -7514064, 5009380,
    -6261722, 6261715,
    crs = osm_crs_native()
  )

  covr1 <- osm_tile_covering(bounds, zoom = osm_zoom_num_tiles(1))
  expect_identical(nrow(covr1), 1L)

  covr4 <- osm_tile_covering(bounds, zoom = osm_zoom_num_tiles(4))
  expect_identical(nrow(covr4), 4L)

  covr16 <- osm_tile_covering(bounds, zoom = osm_zoom_num_tiles(16))
  expect_identical(nrow(covr16), 16L)

  bounds0 <- wk::rct(
    -6261722, 6261715,
    -6261722, 6261715,
    crs = osm_crs_native()
  )

  expect_error(
    osm_tile_covering(bounds0, function(...) FALSE),
    "`zoom` function returned FALSE for all zoom levels"
  )
})

test_that("osm_tile_envelope() works for lng/lat", {
  tiles <- osm_tile(osm_lnglat(-64, 45), 0:4)
  envelopes <- osm_tile_envelope(tiles, wk::wk_crs_longlat())
  expect_identical(wk::wk_crs(envelopes), wk::wk_crs_longlat())

  expect_identical(
    wk::rct_xmin(envelopes),
    c(-180, -180, -90, -90, -67.5)
  )

  expect_identical(
    wk::rct_xmax(envelopes),
    c(180, 0, 0, -45, -45)
  )
})

test_that("osm_tile_envelope() works for native crs", {
  tiles <- osm_tile(osm_lnglat(-64, 45), 0:4)
  envelopes <- osm_tile_envelope(tiles, osm_crs_native())
  expect_identical(wk::wk_crs(envelopes), osm_crs_native())

  top_left_lnglat <- osm_tile_top_left(tiles, wk::wk_crs_longlat())
  top_left_proj <- osm_ensure_native(top_left_lnglat)

  expect_equal(
    wk::rct_xmin(envelopes),
    wk::xy_x(top_left_proj)
  )

  expect_equal(
    wk::rct_ymax(envelopes),
    wk::xy_y(top_left_proj)
  )
})

test_that("osm_tile_envelope() errors for unsupported crs", {
  tiles <- osm_tile(osm_lnglat(-64, 45), 0:4)
  expect_error(
    osm_tile_envelope(tiles, "EPSG:32620"),
    "Unsupported `crs`"
  )
})

test_that("osm_ensure_lnglat() works", {
  expect_error(
    osm_ensure_lnglat(wk::xy(1, 2)),
    "Can't transform NULL"
  )

  expect_identical(
    osm_ensure_lnglat(wk::xy(1, 2, crs = wk::wk_crs_inherit())),
    osm_lnglat(1, 2)
  )

  expect_identical(
    osm_ensure_lnglat(osm_lnglat(1, 2)),
    osm_lnglat(1, 2)
  )

  tiles <- osm_tile(osm_lnglat(-64, 45), 0:4)
  points <- osm_tile_top_left(tiles, crs = wk::wk_crs_longlat())
  points_3857 <- wk::xy(
    c(-20037508.3427892, -20037508.3427892, -10018754.1713946,
      -10018754.1713946, -7514065.62854597),
    c(20037508.3427892, 20037508.3427892, 10018754.1713946,
      10018754.1713946, 7514065.62854597),
    crs = "EPSG:3857"
  )

  # uses the internal formulas with no sf
  expect_equal(osm_ensure_lnglat(points_3857), points)

  skip_if_not_installed("sf")

  # use a crs representation that rosm can't detect as native
  wk::wk_crs(points_3857) <- wk::wk_crs_projjson(sf::st_crs(3857))
  expect_false(crs_is_native(wk::wk_crs(points_3857)))

  expect_equal(osm_ensure_lnglat(points_3857), points)

})

test_that("osm_ensure_native() works", {
  expect_error(
    osm_ensure_native(wk::xy(1, 2)),
    "Can't transform NULL"
  )

  expect_identical(
    osm_ensure_native(wk::xy(1, 2, crs = wk::wk_crs_inherit())),
    osm_native(1, 2)
  )

  expect_identical(
    osm_ensure_native(osm_native(1, 2)),
    osm_native(1, 2)
  )

  tiles <- osm_tile(osm_lnglat(-64, 45), 0:4)
  points <- osm_tile_top_left(tiles, crs = wk::wk_crs_longlat())
  points_3857 <- wk::xy(
    c(-20037508.3427892, -20037508.3427892, -10018754.1713946,
      -10018754.1713946, -7514065.62854597),
    c(20037508.3427892, 20037508.3427892, 10018754.1713946,
      10018754.1713946, 7514065.62854597),
    crs = "EPSG:3857"
  )

  # uses the internal formulas with no sf
  expect_equal(osm_ensure_native(points), points_3857)

  skip_if_not_installed("sf")

  # use a crs representation that rosm can't detect as lnglat
  wk::wk_crs(points) <- wk::wk_crs_projjson(sf::st_crs(4326))
  expect_false(crs_is_lnglat(wk::wk_crs(points)))
  expect_equal(osm_ensure_native(points), points_3857)
})

test_that("ensure_tile() works", {
  expect_error(
    ensure_tile(data.frame()),
    "must have columns `x`, `y`, and `zoom`"
  )

  expect_error(
    ensure_tile(data.frame(x = double(), y = double(), zoom = character())),
    "must be numeric"
  )
})

#' Tile URL Specification
#'
#' See <https://github.com/roblabs/xyz-raster-sources> for a number of useful
#' values to use for `server_url`.
#'
#' @param server_url A url using `\{x\}`, `\{y\}`, and `\{z\}` for the x, y, and
#'   zoom level to be replaced. This can be any URL; non-URLs are assumed to be
#'   local file paths relative to the current working directory at the time of
#'   the download.
#' @param block_size The pixel size of each image
#' @param min_zoom,max_zoom The min/max zoom that this tile specification can handle
#' @param x An object to convert to an osm_url_spec
#' @param content_type A MIME type or NA to guess the type from `server_url`.
#' @param name A name for this spec. Useful for cache specifications.
#' @param ... Passed to S3 methods
#'
#' @return An object of class osm_url_spec.
#' @export
#'
#' @examples
#' osm_url_spec()
#'
osm_url_spec <- function(server_url = "https://tile.openstreetmap.org/{z}/{x}/{y}.png",
                         block_size = c(256, 256),
                         min_zoom = 0,
                         max_zoom = 18,
                         content_type = NA_character_,
                         name = NULL) {
  if (is.null(name)) {
    name <- rlang::hash(server_url)
  }

  stopifnot(
    is.character(server_url), length(server_url) == 1L,
    is.numeric(block_size), length(block_size) == 2L, all(is.finite(block_size)),
    is.numeric(min_zoom), length(min_zoom) == 1L, !is.na(min_zoom),
    is.numeric(max_zoom), length(max_zoom) == 1L, !is.na(max_zoom),
    is.character(content_type), length(content_type) == 1L,
    is.character(name), length(name) == 1L, !is.na(name)
  )

  structure(
    list(
      server_url = server_url,
      block_size = as.integer(block_size),
      min_zoom = as.double(min_zoom),
      max_zoom = as.double(max_zoom),
      content_type = content_type,
      name = name
    ),
    class = "osm_url_spec"
  )
}

#' @rdname osm_url_spec
#' @export
osm_url_spec_example <- function() {
  base <- system.file("extdata/osmns", package = "rosm")
  osm_url_spec(paste0(base, "/{z}_{x}_{y}.png"))
}

#' @rdname osm_url_spec
#' @export
as_osm_url_spec <- function(x, ..., name = NULL) {
  UseMethod("as_osm_url_spec")
}

#' @export
as_osm_url_spec.osm_url_spec <- function(x, ..., name = NULL) {
  if (!is.null(name)) {
    stopifnot(is.character(name), length(name) == 1L, !is.na(name))
    x$name <- name
  }

  x
}

#' @export
as_osm_url_spec.character <- function(x, ..., name = NULL) {
  osm_url_spec(x, name = name)
}

#' Resolve a tile into a URL
#'
#' @param spec An [osm_url_spec()]
#' @inheritParams osm_tile
#'
#' @return A character vector of URLs
#' @export
#'
#' @examples
#' bounds <- wk::rct(
#'   -7514064, 5009380,
#'   -6261722, 6261715,
#'   crs = osm_crs_native()
#' )
#'
#' tiles <- osm_tile_covering(bounds, zoom = 6)
#' osm_url(tiles, osm_url_spec())
#'
osm_url <- function(tile, spec) {
  tile <- osm_tile_normalize(tile)
  n_tile <- nrow(tile)

  spec <- as_osm_url_spec(spec)

  tile <- tile[c("x", "y", "zoom")]
  tile$q <- osm_tile_quadkey(tile)
  names(tile) <- c("x", "y", "z", "q")
  glue_data <- as.environment(tile)
  glue_data$name <- spec$name

  out <- glue::glue_safe(
    spec$server_url,
    .open = "{",
    .close = "}",
    .na = NULL,
    .envir = glue_data
  )

  rep_len(out, n_tile)
}

#' Load tile URLs
#'
#' @inheritParams osm_url
#' @inheritParams osm_tile
#' @param cache_spec An optional [osm_url_spec()] or character vector to be
#'   used as the cache.
#' @param callback A function to be run for each tile fetch or NULL
#'   to do nothing. The callback is always called with two arguments: the first
#'   is the subset of `tile` for which this URL applies (typically one row but
#'   can be more than one in some corner cases); the second is the curl
#'   response object whose useful elements are url, status_code, type, and
#'   content.
#'
#' @return `tile`, invisibly.
#' @export
#'
#' @examples
#' bounds <- wk::rct(
#'   252185, 4815826, 739729, 5210280,
#'   crs = "EPSG:32620"
#' )
#'
#' tiles <- osm_tile_covering(bounds, zoom = 5)
#'
#' osm_url_load_async(
#'   tiles,
#'   osm_url_spec_example(),
#'   function(tile, res) {
#'     str(tile)
#'     str(res)
#'   }
#' )
#'
osm_url_load_async <- function(tile, spec, callback = NULL, cache_spec = NULL) {
  tile <- ensure_tile(tile)
  spec <- as_osm_url_spec(spec)

  if (is.null(cache_spec)) {
    cache_spec <- osm_url_spec(NA_character_)
  } else {
    cache_spec <- as_osm_url_spec(cache_spec, name = spec$name)
  }

  callback <- if (is.null(callback)) function(...) NULL else as.function(callback)

  # calculate the urls and cache values
  tile_url <- osm_url(tile, spec)
  tile_normalized_unique <- which(!duplicated(tile_url) & !is.na(tile_url))

  if (length(tile_normalized_unique) == 0) {
    return(invisible(tile))
  }

  urls <- tile_url[tile_normalized_unique]
  cached <- osm_url(tile[tile_normalized_unique, , drop = FALSE], cache_spec)
  cached <- rep_len(cached, length(tile_normalized_unique))

  # make sure urls are urls (e.g., with file://)
  urls <- ensure_url(urls)

  # make sure cache values are paths
  cached <- ensure_path(cached)

  # replace urls where the cached path exists with a file:// url
  cache_hit <- file.exists(cached)
  cache_hit[is.na(cache_hit)] <- FALSE
  cached_as_url <- ensure_url(cached)
  urls_with_cache <- urls
  urls_with_cache[cache_hit] <- cached_as_url[cache_hit]
  cached[cache_hit] <- NA_character_

  # use curl's async downloader to kick off loads for all tiles in parallel
  # evaluating callback for each as they are completed
  pool <- curl::new_pool(total_con = 6, host_con = 6)
  pb <- progress::progress_bar$new(
    "[:bar]",
    total = length(urls)
  )
  pb$tick(0)
  state <- as.environment(
    list(
      pb = pb,
      tile = tile,
      tile_url = ensure_url(tile_url),
      callback = callback
    )
  )

  for (i in seq_along(urls)) {
    curl::curl_fetch_multi(
      urls_with_cache[i],
      multi_download_async_success(urls[i], cached[i], state),
      multi_download_async_failure(urls[i], cached[i], state),
      pool = pool
    )
  }

  # this will block as long as files are being downloaded
  curl::multi_run(pool = pool)

  invisible(tile)
}

is_url <- function(x) {
  grepl("://", x)
}

ensure_url <- function(x) {
  ifelse(
    is_url(x),
    x,
    paste0("file://", normalizePath(x, winslash = "/", mustWork = FALSE))
  )
}

ensure_path <- function(x) {
  if (any(is_url(x))) {
    stop("Cache results must be paths and not URLs")
  }

  x
}

multi_download_async_success <- function(url, cached, state) {
  force(url)
  force(cached)
  force(state)

  function(res) {
    state$pb$tick()
    tiles <- state$tile[!is.na(state$tile_url) & (state$tile_url == url), , drop = FALSE]
    result <- state$callback(tiles, res)

    # Only write to the cache if the callback succeeds and
    # the callback didn't return FALSE
    if (!is.na(cached) && !identical(result, FALSE)) {
      if (!dir.exists(dirname(cached))) dir.create(dirname(cached), recursive = TRUE)
      con <- file(cached, "wb")
      on.exit(close(con))
      writeBin(res$content, con)
    }
  }
}

multi_download_async_failure <- function(url, cached, state) {
  force(url)
  force(cached)
  force(state)

  function(msg) {
    state$pb$tick()

    res <- structure(
      list(
        url = url,
        status_code = 500,
        msg = msg
      ),
      class = "osm_url_error"
    )

    tiles <- state$tile[!is.na(state$tile_url) & (state$tile_url == url), , drop = FALSE]
    state$callback(tiles, res)
  }
}

test_that("default URL spec works", {
  spec <- osm_url_spec()
  expect_identical(spec$block_size, c(256L, 256L))
  expect_identical(spec$min_zoom, 0)
  expect_identical(spec$max_zoom, 18)
})

test_that("example spec works", {
  spec <- osm_url_spec_example()
  expect_match(spec$server_url, "extdata/osmns/{z}_{x}_{y}.png", fixed = TRUE)
})

test_that("as_osm_url_spec() works", {
  spec <- osm_url_spec()
  expect_identical(as_osm_url_spec(spec), spec)
  expect_identical(as_osm_url_spec(spec$server_url), spec)

  spec_with_default_name <- as_osm_url_spec(spec$server_url, name = "another name")
  expect_identical(
    spec_with_default_name,
    osm_url_spec(name = "another name")
  )

  spec_with_default_name <- as_osm_url_spec(spec, name = "another name")
  expect_identical(
    spec_with_default_name,
    osm_url_spec(name = "another name")
  )
})

test_that("urls can be generated from the default spec", {
  tiles <- data.frame(
    x = c(20, 21, 20, 21),
    y = c(22, 22, 23, 23),
    zoom = 6
  )

  expect_identical(
    osm_url(tiles, osm_url_spec()),
    c(
      "https://tile.openstreetmap.org/6/20/22.png",
      "https://tile.openstreetmap.org/6/21/22.png",
      "https://tile.openstreetmap.org/6/20/23.png",
      "https://tile.openstreetmap.org/6/21/23.png"
    )
  )
})

test_that("urls with quadkeys can be generated", {
  tiles <- data.frame(
    x = c(20, 21, 20, 21),
    y = c(22, 22, 23, 23),
    zoom = 6
  )

  expect_identical(
    osm_url(tiles, "http://something.com/{q}"),
    c(
      "http://something.com/030320",
      "http://something.com/030321",
      "http://something.com/030322",
      "http://something.com/030323"
    )
  )
})

test_that("urls with names can be generated", {
  spec <- osm_url_spec("{name}/{z}/{x}/{y}.png", name = "the_name")
  tiles <- data.frame(x = 1, y = 2, zoom = 3)
  expect_identical(
    osm_url(tiles, spec),
    "the_name/3/1/2.png"
  )
})

test_that("tiles are normalized before generating URLs", {
  oversized_rect <- wk::rct(
    -20038000, -20038000,
    20038000, 20038000,
    crs = osm_crs_native()
  )

  tiles <- osm_tile_covering(oversized_rect, zoom = 0)
  expect_identical(
    unique(as.character(osm_url(tiles, osm_url_spec()))),
    c(NA, "https://tile.openstreetmap.org/0/0/0.png")
  )
})

test_that("url async loader works", {
  tiles <- data.frame(
    x = c(20, 21, 20, 21),
    y = c(22, 22, 23, 23),
    zoom = 6
  )

  spec <- osm_url_spec_example()
  spec$name <- "ex"

  # default does nothing but should write to the cache
  temp_cache <- tempfile()

  cache_spec <- paste0(temp_cache, "/", "{name}_{z}_{x}_{y}.png")
  expect_identical(
    osm_url_load_async(tiles, spec, cache_spec = cache_spec),
    tiles
  )

  expect_setequal(
    list.files(temp_cache),
    c("ex_6_20_22.png", "ex_6_20_23.png", "ex_6_21_22.png", "ex_6_21_23.png")
  )

  unlink(temp_cache, recursive = TRUE)
})

test_that("url async loader can load zero tiles", {
  tiles <- data.frame(x = double(), y = double(), zoom = double())
  expect_identical(osm_url_load_async(tiles, osm_url_spec_example()), tiles)
})

test_that("url async loader errors for paths that are urls", {
  tiles <- data.frame(
    x = c(20, 21, 20, 21),
    y = c(22, 22, 23, 23),
    zoom = 6
  )

  expect_error(
    osm_url_load_async(tiles, osm_url_spec_example(), cache_spec = "http://this.is.a.url"),
    "Cache results must be paths and not URLs"
  )
})

test_that("url async loader runs the error callback", {
  tiles <- data.frame(
    x = c(20, 21, 20, 21),
    y = c(22, 22, 23, 23),
    zoom = 6
  )

  callback <- function(tiles, res) {
    expect_equal(res$status_code, 500)
    stop(glue::glue("<{res$url}> {res$msg}"))
  }

  expect_error(
    osm_url_load_async(tiles, "this_is_not_a_file_anywhere", callback),
    "this_is_not_a_file_anywhere"
  )
})

test_that("url async loader runs the success callback", {
  tiles <- data.frame(
    x = c(20, 21, 20, 21),
    y = c(22, 22, 23, 23),
    zoom = 6
  )

  tiles_out <- data.frame(x = double(), y = double(), zoom = double())

  callback <- function(tiles, res) {
    expect_identical(res$status_code, 0L)
    expect_identical(nrow(tiles), 1L)
    tiles_out <<- rbind(tiles_out, tiles)
  }

  expect_identical(
    osm_url_load_async(tiles, osm_url_spec_example(), callback = callback),
    tiles
  )

  expect_identical(tiles_out[order(tiles_out$y, tiles_out$x), ], tiles)
})

test_that("url async loader stops for callback error", {
  tiles <- data.frame(
    x = c(20, 21, 20, 21),
    y = c(22, 22, 23, 23),
    zoom = 6
  )

  # After a callback error, nothing should be cached
  temp_cache <- tempfile()
  cache_spec <- paste0(temp_cache, "/", "{z}_{x}_{y}.png")

  callback <- function(tiles, res) {
    stop("In the name of Open Street Map!")
  }

  expect_error(
    osm_url_load_async(tiles, osm_url_spec_example(), callback = callback),
    "In the name of Open Street Map!"
  )

  expect_false(dir.exists(temp_cache))
})

#' Load an Open Street Map image
#'
#' @inheritParams osm_url_load_async
#' @inheritParams osm_tile_covering
#' @param quiet Use `TRUE` for fewer messages or `FALSE` for more messages.
#'
#' @return A [wk::grd_rct()] whose data member is a nativeRaster.
#' @export
#'
#' @examples
#' bounds <- wk::rct(
#'   -7476083, 5349058,
#'   -6594103, 6243203,
#'   crs = osm_crs_native()
#' )
#'
#' (grd <- osm_raster(bounds, osm_url_spec()))
#' plot(grd)
#'
osm_raster <- function(bbox, spec, zoom = osm_zoom_num_tiles(6),
                       cache_spec = NULL, quiet = NA) {
  spec <- as_osm_url_spec(spec)
  if (is.na(spec$content_type)) {
    content_type <- guess_content_type(spec$server_url)
  } else {
    content_type <- spec$content_type
  }

  loader <- get_native_raster_loader(content_type)

  tiles <- osm_tile_covering(bbox, zoom)

  if (nrow(tiles) == 0) {
    empty_native_raster <- structure(integer(), class = "nativeRaster")
    return(wk::grd_rct(matrix(empty_native_raster, nrow = 0, ncol = 0), bbox))
  }

  x_tile_range <- range(tiles$x)
  y_tile_range <- range(tiles$y)
  tiles_x <- diff(x_tile_range) + 1L
  tiles_y <- diff(y_tile_range) + 1L
  pixels_x <- spec$block_size[1] * tiles_x
  pixels_y <- spec$block_size[2] * tiles_y

  state <- new.env(parent = emptyenv())
  state$template <- matrix(
    integer(pixels_x * pixels_y),
    ncol = pixels_x,
    nrow = pixels_y
  )

  state$n <- 0L
  state$n_loaded <- 0L
  state$range_x <- x_tile_range
  state$range_y <- y_tile_range
  state$block_size <- spec$block_size
  state$loader <- loader
  state$quiet <- !identical(quiet, FALSE)

  callback <- osm_raster_callback(state)
  osm_url_load_async(
    tiles,
    spec,
    callback,
    cache_spec = cache_spec
  )

  if (state$n_loaded == 0 && !isTRUE(quiet)) {
    stop(
      glue::glue("0/{state$n} tiles loaded. Use `quiet = FALSE` to see error messages")
    )
  } else if (state$n_loaded != state$n && !isTRUE(quiet)) {
    warning(
      glue::glue("{state$n_loaded}/{state$n} tiles loaded.")
    )
  }

  tile_bbox <- wk::wk_bbox(osm_tile_envelope(tiles))
  class(state$template) <- "nativeRaster"
  wk::grd_rct(state$template, tile_bbox)
}

osm_raster_callback <- function(state) {
  force(state)

  function(tiles, res) {
    # Keep track of loaded/not loaded so we can error if everything fails
    state$n <- state$n + 1L

    # Errors aren't the end of the world here...some providers return a 404
    # if a tile should be blank. file:// urls will have a status code of 0
    # on success but will go through the error callback if the file isn't found,
    # hence trying to catch that here to simulate a 404.
    if (inherits(res, "osm_url_error") && grepl("Couldn't open file", res$msg)) {
      res$status_code <- 404
    } else if (inherits(res, "osm_url_error")) {
      if (!state$quiet) message(sprintf("[%s] %s\n %s", res$status_code, res$url, res$msg))
      return(FALSE)
    }

    if (res$status_code >= 300) {
      if (!state$quiet) message(sprintf("[%s] %s", res$status_code, res$url))
      return(FALSE)
    }

    if (!state$quiet) message(sprintf("[200] %s", res$url))

    native <- try(state$loader(res$content), silent = state$quiet)
    if (inherits(native, "try-error")) {
      return(FALSE)
    }

    # Keep track of loaded/not loaded so we can error if everything fails
    state$n_loaded <- state$n_loaded + 1L

    # The indexing is tricky here because nativeRaster is actually row-major
    # despite its dimensions and subset/subset-assign methods lying about it
    ix0 <- tiles$x - state$range_x[1]
    iy0 <- tiles$y - state$range_y[1]
    ipxx0 <- (ix0 * state$block_size[1])
    ipxy0 <- (iy0 * state$block_size[2])
    ipxx <- rep(ipxx0:(ipxx0 + state$block_size[1] - 1), state$block_size[2])
    ipxy <- rep(ipxy0:(ipxy0 + state$block_size[2] - 1), each = state$block_size[1])
    ipx <- ipxy * ncol(state$template) + ipxx + 1L

    state$template[ipx] <- native
  }
}

guess_content_type <- function(server_url) {
  before_query_string <- strsplit(server_url, "?", fixed = TRUE)[[1]][1]
  ext <- tools::file_ext(before_query_string)
  switch(
    ext,
    png = "image/png",
    jpg = "image/jpeg",
    jpeg = "image/jpeg",
    tif = "image/tiff",
    tiff = "image/tiff",
    stop(sprintf("Can't guess content type from server url '%s'", server_url))
  )
}

get_native_raster_loader <- function(content_type) {
  switch(
    content_type,
    "image/png" = function(x) png::readPNG(x, native = TRUE),
    "image/jpeg" = function(x) jpeg::readJPEG(x, native = TRUE),
    "image/tiff" = function(x) tiff::readTIFF(x, native = TRUE),
    stop(sprintf("Can't guess nativeRaster loader from content type '%s'", content_type))
  )
}

test_that("osm_raster produces errors and warnings when appropriate", {
  bounds <- wk::rct(
    -7514064, 5009380,
    -6261722, 6261715,
    crs = osm_crs_native()
  )

  expect_error(
    expect_message(
      osm_raster(bounds, "fileXX:///bad_url.png", quiet = NA),
      "^\\[500\\] fileXX"
    ),
    "0/1 tiles loaded"
  )

  expect_error(
    expect_message(
      osm_raster(bounds, "fileXX:///bad_url.png", quiet = FALSE),
      "^\\[500\\] fileXX"
    ),
    "0/1 tiles loaded"
  )

  expect_silent(osm_raster(bounds, "fileXX:///bad_url.png", quiet = TRUE))

  expect_error(
    expect_message(
      osm_raster(bounds, "file:///does_not_exist.png", quiet = NA),
      "^\\[404\\] file://"
    ),
    "0/1 tiles loaded"
  )

  expect_warning(
    osm_raster(bounds, osm_url_spec_example(), quiet = NA),
    "9/16 tiles loaded"
  )
})

test_that("osm_raster() produces the correct raster image", {
  bounds <- wk::rct(
    -7476083, 5349058,
    -6594103, 6243203,
    crs = osm_crs_native()
  )

  grd <- osm_raster(bounds, osm_url_spec_example())
  expect_s3_class(grd, "wk_grd")
  expect_identical(wk::wk_crs(grd), osm_crs_native())

  vdiffr::expect_doppelganger("basic-osm-raster", plot(grd))
})

test_that("guess_content_type() works", {
  expect_identical(guess_content_type("xx.png"), "image/png")
  expect_identical(guess_content_type("xx.jpg"), "image/jpeg")
  expect_identical(guess_content_type("xx.jpeg"), "image/jpeg")
  expect_identical(guess_content_type("xx.tif"), "image/tiff")
  expect_identical(guess_content_type("xx.tiff"), "image/tiff")
  expect_error(guess_content_type("xx.exe"), "Can't guess content type")
})

test_that("get_native_raster_loader() works", {
  expect_type(get_native_raster_loader("image/png"), "closure")
  expect_type(get_native_raster_loader("image/jpeg"), "closure")
  expect_type(get_native_raster_loader("image/tiff"), "closure")
  expect_error(
    get_native_raster_loader("not supported"),
    "Can't guess nativeRaster loader"
  )
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/url.R
\name{osm_url_spec}
\alias{osm_url_spec}
\alias{osm_url_spec_example}
\alias{as_osm_url_spec}
\title{Tile URL Specification}
\usage{
osm_url_spec(
  server_url = "https://tile.openstreetmap.org/{z}/{x}/{y}.png",
  block_size = c(256, 256),
  min_zoom = 0,
  max_zoom = 18,
  content_type = NA_character_,
  name = NULL
)

osm_url_spec_example()

as_osm_url_spec(x, ..., name = NULL)
}
\arguments{
\item{server_url}{A url using `\{x\}`, `\{y\}`, and `\{z\}` for the x, y, and
zoom level to be replaced. This can be any URL; non-URLs are assumed to be
local file paths relative to the current working directory at the time of
the download.}

\item{block_size}{The pixel size of each image}

\item{min_zoom, max_zoom}{The min/max zoom that this tile specification can handle}

\item{content_type}{A MIME type or NA to guess the type from `server_url`.}

\item{name}{A name for this spec. Useful for cache specifications.}

\item{x}{An object to convert to an osm_url_spec}

\item{...}{Passed to S3 methods}
}
\value{
An object of class osm_url_spec.
}
\description{
See <https://github.com/roblabs/xyz-raster-sources> for a number of useful
values to use for `server_url`.
}
\examples{
osm_url_spec()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tile-math.R
\name{osm_tile}
\alias{osm_tile}
\alias{osm_tile_quadkey}
\alias{osm_tile_top_left}
\alias{osm_tile_envelope}
\title{Low-level tile math}
\usage{
osm_tile(pt, zoom)

osm_tile_quadkey(tile)

osm_tile_top_left(tile, crs = osm_crs_native())

osm_tile_envelope(tile, crs = osm_crs_native())
}
\arguments{
\item{pt}{A vector of points as coerced by [wk::as_xy()]. The CRS for these
points is considered.}

\item{zoom}{A zoom level, generally between 0 and 21, with higher values
representing a smaller (i.e., more detailed) tile.}

\item{tile}{A `data.frame()` with columns `x`, `y`, and `zoom`.}

\item{crs}{A target CRS. Either [wk::wk_crs_longlat()] or
[osm_crs_native()].}
}
\value{
- `osm_tile()`: A `data.frame()` with columns `x`, `y`, and `zoom`.
  - `osm_tile_top_left()`: A [wk::xy()] of the top-left (northwest) corner
    of the tile.
  - `osm_tile_envelope()`: A [wk::rct()] of the tile bounds.
}
\description{
Open Street Map operates using a system of tiles whose value and bounds
are easily calculated from WGS84 longitude/latitude. These functions convert
between tile system coordinates and longitude/latitude.
}
\examples{
(tiles <- osm_tile(osm_lnglat(-64, 45), zoom = 0:5))
osm_tile_envelope(tiles)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/url.R
\name{osm_url}
\alias{osm_url}
\title{Resolve a tile into a URL}
\usage{
osm_url(tile, spec)
}
\arguments{
\item{tile}{A `data.frame()` with columns `x`, `y`, and `zoom`.}

\item{spec}{An [osm_url_spec()]}
}
\value{
A character vector of URLs
}
\description{
Resolve a tile into a URL
}
\examples{
bounds <- wk::rct(
  -7514064, 5009380,
  -6261722, 6261715,
  crs = osm_crs_native()
)

tiles <- osm_tile_covering(bounds, zoom = 6)
osm_url(tiles, osm_url_spec())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{has_internet}
\alias{has_internet}
\title{Check for Internet}
\usage{
has_internet()
}
\value{
TRUE if the internet is available, false otherwise
}
\description{
Used to skip tests and examples for this package when offline.
}
\examples{
has_internet()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/url.R
\name{osm_url_load_async}
\alias{osm_url_load_async}
\title{Load tile URLs}
\usage{
osm_url_load_async(tile, spec, callback = NULL, cache_spec = NULL)
}
\arguments{
\item{tile}{A `data.frame()` with columns `x`, `y`, and `zoom`.}

\item{spec}{An [osm_url_spec()]}

\item{callback}{A function to be run for each tile fetch or NULL
to do nothing. The callback is always called with two arguments: the first
is the subset of `tile` for which this URL applies (typically one row but
can be more than one in some corner cases); the second is the curl
response object whose useful elements are url, status_code, type, and
content.}

\item{cache_spec}{An optional [osm_url_spec()] or character vector to be
used as the cache.}
}
\value{
`tile`, invisibly.
}
\description{
Load tile URLs
}
\examples{
bounds <- wk::rct(
  252185, 4815826, 739729, 5210280,
  crs = "EPSG:32620"
)

tiles <- osm_tile_covering(bounds, zoom = 5)

osm_url_load_async(
  tiles,
  osm_url_spec_example(),
  function(tile, res) {
    str(tile)
    str(res)
  }
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{osm_raster}
\alias{osm_raster}
\title{Load an Open Street Map image}
\usage{
osm_raster(
  bbox,
  spec,
  zoom = osm_zoom_num_tiles(6),
  cache_spec = NULL,
  quiet = NA
)
}
\arguments{
\item{bbox}{A [wk::rct()] or object with a [wk::wk_bbox()] method.}

\item{spec}{An [osm_url_spec()]}

\item{zoom}{A zoom level or an auto zoom specifier like
[osm_zoom_num_tiles()].}

\item{cache_spec}{An optional [osm_url_spec()] or character vector to be
used as the cache.}

\item{quiet}{Use `TRUE` for fewer messages or `FALSE` for more messages.}
}
\value{
A [wk::grd_rct()] whose data member is a nativeRaster.
}
\description{
Load an Open Street Map image
}
\examples{
bounds <- wk::rct(
  -7476083, 5349058,
  -6594103, 6243203,
  crs = osm_crs_native()
)

(grd <- osm_raster(bounds, osm_url_spec()))
plot(grd)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{set_default_cachedir}
\alias{set_default_cachedir}
\alias{get_default_cachedir}
\title{Set/Get the Default Tile Cache Location}
\usage{
set_default_cachedir(cachedir)

get_default_cachedir()
}
\arguments{
\item{cachedir}{A path to use as the cache directory (relative to the working directory).
Use NULL to reset to the default.}
}
\value{
The previous cache directory, invisibly.
}
\description{
The default tile cache location is the "rosm.cache" folder in the current
working directory, but for a variety of reasons it may be desirable to
use one cache directory for all calls in a script. This must be called
every time the namespace is loaded.
}
\examples{
set_default_cachedir(tempfile())
get_default_cachedir()
(set_default_cachedir(NULL))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tile-math.R
\name{osm_native}
\alias{osm_native}
\alias{osm_lnglat}
\alias{osm_crs_native}
\alias{osm_ensure_lnglat}
\alias{osm_ensure_native}
\title{Coordinate helpers}
\usage{
osm_native(x, y)

osm_lnglat(lng, lat)

osm_crs_native()

osm_ensure_lnglat(pt)

osm_ensure_native(pt)
}
\arguments{
\item{x, y}{Ordinate values in EPSG:3857 (Spherical Mercator in meters)}

\item{lng, lat}{Coordinate values for longitude/latitude in degrees.}

\item{pt}{A vector of points as coerced by [wk::as_xy()]. The CRS for these
points is considered.}
}
\value{
- `osm_native()`, `osm_lnglat()`, `osm_ensure_native()`, and
    `osm_ensure_lnglat()` return a [wk::xy()] with the appropriate
    crs
  - `osm_crs_native()` returns a value that can be used as the [wk::wk_crs()]
    of a vector.
}
\description{
Coordinate helpers
}
\examples{
osm_lnglat(-64, 45)
osm_ensure_native(osm_lnglat(-64, 45))
osm_ensure_lnglat(
  osm_ensure_native(osm_lnglat(-64, 45))
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{as.tile_source}
\alias{as.tile_source}
\alias{is.tile_source}
\alias{source_from_url_format}
\alias{register_tile_source}
\alias{set_default_tile_source}
\alias{get_default_tile_source}
\alias{osm.types}
\alias{bmaps.types}
\alias{bmaps.plot}
\alias{extract_bbox}
\alias{osm.plot}
\alias{osm.image}
\alias{osm.raster}
\alias{osm.points}
\alias{osm.segments}
\alias{osm.lines}
\alias{osm.polygon}
\alias{osm.text}
\alias{makebbox}
\alias{zoombbox}
\title{Deprecated interface}
\usage{
as.tile_source(x, ..., api_key = NULL)

is.tile_source(x)

source_from_url_format(
  url_format,
  max_zoom = tile.maxzoom.default(),
  min_zoom = 0,
  attribution = NULL,
  extension = tools::file_ext(url_format[1]),
  ...
)

register_tile_source(...)

set_default_tile_source(x, ...)

get_default_tile_source()

osm.types()

bmaps.types()

bmaps.plot(bbox, type = "Aerial", key = NULL, ...)

extract_bbox(x, tolatlon = TRUE, ...)

osm.plot(
  bbox,
  zoomin = 0,
  zoom = NULL,
  type = NULL,
  forcedownload = FALSE,
  stoponlargerequest = TRUE,
  fusetiles = TRUE,
  cachedir = NULL,
  res = 150,
  project = TRUE,
  progress = c("text", "none"),
  quiet = TRUE,
  api_key = NULL,
  ...
)

osm.image(
  x,
  zoomin = 0,
  zoom = NULL,
  type = NULL,
  forcedownload = FALSE,
  cachedir = NULL,
  progress = c("text", "none"),
  quiet = TRUE,
  api_key = NULL
)

osm.raster(
  x,
  zoomin = 0,
  zoom = NULL,
  type = "osm",
  forcedownload = FALSE,
  cachedir = NULL,
  progress = c("text", "none"),
  quiet = TRUE,
  projection = NULL,
  crop = FALSE,
  filename = NULL,
  resample = "bilinear",
  api_key = NULL,
  ...
)

osm.points(x, y = NULL, epsg = 4326, toepsg = 3857, ...)

osm.segments(x0, y0, x1 = x0, y1 = y0, epsg = 4326, toepsg = 3857, ...)

osm.lines(x, y = NULL, epsg = 4326, toepsg = 3857, ...)

osm.polygon(x, y = NULL, epsg = 4326, toepsg = 3857, ...)

osm.text(x, y = NULL, labels = seq_along(x), epsg = 4326, toepsg = 3857, ...)

makebbox(n, e, s, w)

zoombbox(bbox, factor = 1, offset = c(0, 0))
}
\arguments{
\item{x, y, x0, y0, x1, y1, url_format, max_zoom, min_zoom, attribution, extension}{Deprecated}

\item{...}{Arguments passed to other methods}

\item{api_key}{API key used for CARTO map types. If `NULL`, the value of
the `CARTO_API_KEY` environment variable is used.}

\item{bbox}{A bounding box as generated by \code{sp::bbox()}}

\item{type}{A map type; one of that returned by \link{osm.types}. User defined types are possible
by defining \code{tile.url.TYPENAME <- function(xtile, ytile, zoom){}} and passing TYPENAME
as the \code{type} argument.}

\item{key, tolatlon, epsg, toepsg, labels, n, e, s, w, factor, offset}{Deprecated}

\item{zoomin}{The amount by which to adjust the automatically calculated zoom (or
manually specified if the \code{zoom} parameter is passed). Use +1 to zoom in, or -1 to zoom out.}

\item{zoom}{Manually specify the zoom level (not reccomended; adjust \code{zoomin} or
\code{res} instead.}

\item{forcedownload}{\code{TRUE} if cached tiles should be re-downloaded. Useful if
some tiles are corrupted.}

\item{stoponlargerequest}{By default \code{osm.plot} will only load 32 tiles at a time. If
plotting at a higher resolution it may be necessary to pass \code{true} here.}

\item{fusetiles}{\code{TRUE} if tiles should be fused into a single image. This is the
default because white lines appear between tiles if it is set to \code{FALSE}. PDFs
appear not to have this problem, so when plotting large, high resolution PDFs it may be
faster (and more memory efficient) to use \code{fusetiles=FALSE}.}

\item{cachedir}{The directory in which tiles should be cached. Defaults to \code{getwd()/rosm.cache}.}

\item{res}{The resolution used to calculate scale.}

\item{project}{\code{TRUE} if tiles should be projected to a pseudo-mercator projection,
\code{FALSE} if lat/lon should be maintained. Becuase \code{sp::plot} adjusts the aspect
according to latitude for lat/lon coordinates, this makes little difference at high
zoom and may make plotting overlays more convenient. Defaults to \code{TRUE}.}

\item{progress}{A progress bar to use, or "none" to suppress progress updates}

\item{quiet}{Pass \code{FALSE} to see more error messages, particularly if
your tiles do not download/load properly.}

\item{projection}{A map projection in which to reproject the RasterStack as
generated by \code{CRS()} or \code{Spatial*@proj4string}. If a
\code{Spatial*} object is passed as the first argument, this argument will
be ignored.}

\item{crop}{\code{TRUE} if results should be cropped to the specified
bounding box (see \code{x}), \code{FALSE} otherwise.}

\item{filename}{A filename to which the raster should be written (see
\code{raster::writeRaster()}). Use a ".tif" extension to write as a
GeoTIFF.}

\item{resample}{One of "ngb" (nearest neighbour) or "bilinear". Passed to
\link[raster]{projectRaster}.}
}
\description{
The previous interface for rosm was written to support idioms that are
no longer prevalent in modren r-spatial code. These functions may continue
to exist; however, their use is not encouraged and the functions may be
removed in a future release.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tile-math.R
\name{osm_tile_covering}
\alias{osm_tile_covering}
\alias{osm_zoom_num_tiles}
\title{Get an OSM tile covering}
\usage{
osm_tile_covering(bbox, zoom = osm_zoom_num_tiles(6))

osm_zoom_num_tiles(num_tiles)
}
\arguments{
\item{bbox}{A [wk::rct()] or object with a [wk::wk_bbox()] method.}

\item{zoom}{A zoom level or an auto zoom specifier like
[osm_zoom_num_tiles()].}

\item{num_tiles}{The minimum number of tiles to use when choosing a
zoom level.}
}
\value{
- `osm_tile_covering()` returns a `data.frame()` with columns x, y, and zoom.
}
\description{
Get an OSM tile covering
}
\examples{
bounds <- wk::rct(
  -7514064, 5009380,
  -6261722, 6261715,
  crs = osm_crs_native()
)

osm_tile_covering(bounds)

}
